"""
Combine the AnalogSignal and Event objects from different wave analysis blocks
into the same Neo Block.
"""

import argparse
from pathlib import Path
from utils.io_utils import load_neo, write_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--properties", nargs='*', type=Path, default=[],
                 help="paths to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    waves_block = load_neo(args.data)

    asig_names = [asig.name for asig in waves_block.segments[0].analogsignals]
    event_names = [event.name for event in waves_block.segments[0].events]

    if not args.properties or not args.properties[0]:
        args.properties = []

    for property in args.properties:
        block = load_neo(property)

        for asig in block.segments[0].analogsignals:
            if asig.name not in asig_names:
                waves_block.segments[0].analogsignals.append(asig)

        for event in block.segments[0].events:
            if event.name in event_names:
                waves_evt = waves_block.filter(name=event.name, objects="Event")[0]
                for key, value in event.annotations.items():
                    if key not in waves_evt.annotations:
                        waves_evt.annotations[key] = value
                for key, value in event.array_annotations.items():
                    if key not in waves_evt.array_annotations:
                        waves_evt.array_annotations[key] = value
            else:
                waves_block.segments[0].events.append(event)

        del block

    write_neo(args.output, waves_block)



"""
Calculate the optical flow (vector field) signal 
using the Horn Schunck algorithm.
"""

import argparse
from scipy.ndimage import gaussian_filter
from scipy.signal import hilbert
import neo
import numpy as np
from copy import copy
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_str, str_to_bool
from utils.neo_utils import imagesequence_to_analogsignal, analogsignal_to_imagesequence
from utils.convolve import phase_conv2d, get_kernel, conv, norm_angle

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--alpha", nargs='?', type=float, default=0.001,
                 help='regularization parameter')
CLI.add_argument("--max_Niter", nargs='?', type=int, default=50,
                 help='maximum number of iterations')
CLI.add_argument("--convergence_limit", nargs='?', type=float, default=10e-3,
                 help='absolute change at which consider optimization converged')
CLI.add_argument("--gaussian_sigma", nargs='+', type=float, default=[0,3,3],
                 help='sigma of gaussian filter in each dimension')
CLI.add_argument("--derivative_filter", nargs='?', type=none_or_str, default=None,
                 help='Filter kernel to use for calculating spatial derivatives')
CLI.add_argument("--use_phases", nargs='?', type=str_to_bool, default=False,
                 help='whether to use signal phase instead of amplitude')

def horn_schunck_step(frame, next_frame, alpha, max_Niter, convergence_limit,
                      kernelHS, kernelT, kernelX, kernelY,
                      are_phases=False, kernel_center=None):
    """
    Parameters
    ----------
    frame: numpy.ndarray
        image at t=0
    next_frame: numpy.ndarray
        image at t=1
    alpha: float
        regularization constant
    max_Niter: int
        maximum number of iteration
    convergence_limit: float
        the maximum absolute change between iterations defining convergence
    """
    # set up initial velocities
    vx = np.zeros_like(frame)
    vy = np.zeros_like(frame)

    # estimate derivatives
    [fx, fy, ft] = compute_derivatives(frame, next_frame,
                                       kernelX=kernelX,
                                       kernelY=kernelY,
                                       kernelT=kernelT,
                                       are_phases=are_phases)
    # iteration to reduce error
    for i in range(max_Niter):
        # Compute local averages of the flow vectors (smoothness constraint)
        vx_avg = conv(vx, kernelHS)
        vy_avg = conv(vy, kernelHS)
        # common part of update step (brightness constancy)
        der = (fx*vx_avg + fy*vy_avg + ft) / (alpha**2 + fx**2 + fy**2)
        # iterative step
        new_vx = vx_avg - fx * der
        new_vy = vy_avg - fy * der
        # check convergence
        max_dv = np.max([np.max(np.abs(vx-new_vx)),
                         np.max(np.abs(vy-new_vy))])
        vx, vy = new_vx, new_vy
        if max_dv < convergence_limit:
            break
    return vx + vy*1j


def compute_derivatives(frame, next_frame, kernelX, kernelY,
                        kernelT, are_phases=False):
    if are_phases:
        fx = phase_conv2d(frame, kernelX) \
           + phase_conv2d(next_frame, kernelX)
        fy = phase_conv2d(frame, kernelY) \
           + phase_conv2d(next_frame, kernelY)
        ft = norm_angle(frame - next_frame)
    else:
        fx = conv(frame, kernelX) + conv(next_frame, kernelX)
        fy = conv(frame, kernelY) + conv(next_frame, kernelY)
        # ft = conv(frame, kernelT) + conv(next_frame, -kernelT)
        ft = frame - next_frame
    return fx, fy, ft

def horn_schunck(frames, alpha, max_Niter, convergence_limit,
                 kernelHS, kernelT, kernelX, kernelY,
                 are_phases=False):

    nan_channels = np.where(np.bitwise_not(np.isfinite(frames[0])))
    frames = interpolate_empty_sites(frames, are_phases)

    vector_frames = np.zeros(frames.shape, dtype=complex)

    for i, frame in enumerate(frames[:-1]):
        next_frame = frames[i+1]

        vector_frames[i] = horn_schunck_step(frame,
                                             next_frame,
                                             alpha=alpha,
                                             max_Niter=max_Niter,
                                             convergence_limit=convergence_limit,
                                             kernelHS=kernelHS,
                                             kernelT=kernelT,
                                             kernelX=kernelX,
                                             kernelY=kernelY,
                                             are_phases=are_phases)
        vector_frames[i][nan_channels] = np.nan + np.nan*1j

    frames[:,nan_channels[0],nan_channels[1]] = np.nan
    return vector_frames


def interpolate_empty_sites(frames, are_phases=False):
    if np.isfinite(frames).all():
        return frames
    dim_y, dim_x = frames[0].shape
    grid_y, grid_x = np.meshgrid([-1,0,1],[-1,0,1], indexing='ij')

    for i, frame in enumerate(frames):
        new_frame = copy(frame)
        while not np.isfinite(new_frame).all():
            y, x = np.where(np.bitwise_not(np.isfinite(new_frame)))
            # loop over nan-sites
            for xi, yi in zip(x,y):
                neighbours = []
                # collect neighbours of each site
                for dx, dy in zip(grid_x.flatten(), grid_y.flatten()):
                    xn = xi+dx
                    yn = yi+dy
                    if (0 <= xn) & (xn < dim_x) & (0 <= yn) & (yn < dim_y):
                        neighbours.append(frames[i, yn, xn])
                # average over neihbour values
                if np.isfinite(neighbours).any():
                    if are_phases:
                        vectors = np.exp(1j*np.array(neighbours))
                        new_frame[yi,xi] = np.angle(np.nansum(vectors))
                    else:
                        new_frame[yi,xi] = np.nansum(neighbours)
            frames[i] = new_frame
    return frames


def smooth_frames(frames, sigma):
    # replace nan sites by median
    if np.isfinite(frames).any():
        # assume constant nan sites over time
        nan_sites = np.where(np.bitwise_not(np.isfinite(frames[0])))
        if np.iscomplexobj(frames):
            frames[:,nan_sites[0],nan_sites[1]] = np.nanmedian(np.real(frames)) \
                                                + np.nanmedian(np.imag(frames))*1j
        else:
            frames[:,nan_sites[0],nan_sites[1]] = np.nanmedian(frames)
    else:
        nan_sites = None

    # apply gaussian filter
    if np.iscomplexobj(frames):
        frames = gaussian_filter(np.real(frames), sigma=sigma, mode='nearest') \
               + gaussian_filter(np.imag(frames), sigma=sigma, mode='nearest')*1j
    else:
        frames = gaussian_filter(frames, sigma=sigma, mode='nearest')

    # set nan sites back to nan
    if nan_sites is not None:
        if np.iscomplexobj(frames):
            frames[:,nan_sites[0],nan_sites[1]] = np.nan + np.nan*1j
        else:
            frames[:,nan_sites[0],nan_sites[1]] = np.nan

    return frames


def plot_opticalflow(frame, vec_frame, skip_step=None, are_phases=False):
    # Every <skip_step> point in each direction.
    fig, ax = plt.subplots()
    dim_y, dim_x = vec_frame.shape
    if are_phases:
        cmap = 'twilight'
        vmin, vmax = -np.pi, np.pi
    else:
        cmap = 'viridis'
        vmin, vmax = None, None
    img = ax.imshow(frame, interpolation='nearest', vmin=vmin, vmax=vmax,
                    cmap=plt.get_cmap(cmap), origin='lower')
    plt.colorbar(img, ax=ax)
    if skip_step is None:
        skip_step = int(min([dim_x, dim_y]) / 50) + 1

    ax.quiver(np.arange(dim_x)[::skip_step],
              np.arange(dim_y)[::skip_step],
              np.real(vec_frame[::skip_step,::skip_step]),
              np.imag(vec_frame[::skip_step,::skip_step]))

    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    return ax

def is_phase_signal(signal, use_phases):
    vmin = np.nanmin(signal)
    vmax = np.nanmax(signal)
    in_range = np.isclose(np.array([vmin,   vmax]),
                          np.array([-np.pi, np.pi]), atol=0.05)
    if in_range.all():
        print('The signal values seem to be phase values [-pi, pi]!')
        print(f'The setting "use_phases" is {bool(use_phases)}.')
        if use_phases:
            print('Thus, the phase transformation is skipped.')
        else:
            print('Anyhow, the signal is treated as phase signal in the following. If this is not desired please review the preprocessing.')
        return True
    else:
        return False


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)

    frames = imgseq.as_array()
    # frames /= np.nanmax(np.abs(frames))

    if is_phase_signal(frames, args.use_phases):
        args.use_phases = True
    elif args.use_phases:
        analytic_frames = hilbert(frames, axis=0)
        frames = np.angle(analytic_frames)

    kernelHS = np.array([[1, 2, 1],
                         [2, 0, 2],
                         [1, 2, 1]], dtype=float) * 1/12
    kernel = get_kernel(args.derivative_filter)
    kernelT = np.ones_like(kernel.x, dtype=float)
    kernelT /= np.sum(kernelT)

    vector_frames = horn_schunck(frames=frames,
                                 alpha=args.alpha,
                                 max_Niter=args.max_Niter,
                                 convergence_limit=args.convergence_limit,
                                 kernelX=kernel.x,
                                 kernelY=kernel.y,
                                 kernelT=kernelT,
                                 kernelHS=kernelHS,
                                 are_phases=args.use_phases)

    if np.sum(args.gaussian_sigma):
        vector_frames = smooth_frames(vector_frames, sigma=args.gaussian_sigma)

    vec_imgseq = neo.ImageSequence(vector_frames,
                                   units='dimensionless',
                                   dtype=complex,
                                   t_start=imgseq.t_start,
                                   spatial_scale=imgseq.spatial_scale,
                                   sampling_rate=imgseq.sampling_rate,
                                   name='optical_flow',
                                   description='Horn-Schunck estimation of optical flow',
                                   file_origin=imgseq.file_origin)
   
    vec_imgseq.annotations = copy(imgseq.annotations)

    if args.output_img is not None:
        ax = plot_opticalflow(frames[10], vector_frames[10],
                              skip_step=None, are_phases=args.use_phases)
        ax.set_ylabel(f'pixel size: {imgseq.spatial_scale} ')
        ax.set_xlabel('{:.3f} s'.format(asig.times[0].rescale('s').magnitude))
        save_plot(args.output_img)

    # block.segments[0].imagesequences = [vec_imgseq]
    vec_asig = imagesequence_to_analogsignal(vec_imgseq)

    block.segments[0].analogsignals.append(vec_asig)

    write_neo(args.output, block)



"""
Check whether the input data representation adheres to the stage's requirements.

Additionally prints a short summary of the data attributes.
"""

import numpy as np
import argparse
import quantities as pq
from utils.io_utils import load_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if len(block.segments) > 1:
        print("More than one Segment found; all except the first one " \
            + "will be ignored.")
    if len(block.segments[0].analogsignals) > 1:
        print("More than one AnalogSignal found; all except the first one " \
            + "will be ignored.")

    asig = block.segments[0].analogsignals[0]

    x_coords = asig.array_annotations['x_coords']
    y_coords = asig.array_annotations['y_coords']

    print('Recording Time:\t\t', asig.t_stop - asig.t_start)
    print('Sampling Rate:\t\t', asig.sampling_rate)
    num_channels = np.count_nonzero(~np.isnan(np.sum(asig, axis=0)))
    print('Number of Channels:\t', num_channels)

    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1
    print('Grid Dimensions:\t', f'{dim_x} x {dim_y}')
    print('Empty Grid Sites:\t', dim_x*dim_y - num_channels)



"""
Calculate the wave directions per wave and channel, 
based on the spatial gradient of wave trigger times.
"""

import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to spatial derivative dataframe")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    df = pd.read_csv(args.data)

    direction_df = pd.DataFrame(df.channel_id, columns=['channel_id'])
    direction_df['direction_local_x'] = df.dt_x
    direction_df['direction_local_y'] = df.dt_y
    direction_df[f'{args.event_name}_id'] = df[f'{args.event_name}_id']

    direction_df.to_csv(args.output)

    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    ax.hist(np.angle(df.dt_x + 1j*df.dt_y), bins=36, range=[-np.pi, np.pi])

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Downsample the input data by combining and averaging neighboring channels.
"""

import numpy as np
import matplotlib.pyplot as plt
import argparse
import os
import neo
from skimage import measure
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_str
from utils.neo_utils import analogsignal_to_imagesequence, imagesequence_to_analogsignal

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str,
                 help="path of output image", default=None)
CLI.add_argument("--macro_pixel_dim", nargs='?', type=int,
                 help="smoothing factor", default=2)

def spatial_smoothing(imgseq, macro_pixel_dim):
    images_reduced = measure.block_reduce(imgseq.as_array(),
                                          block_size=(1, macro_pixel_dim, macro_pixel_dim),
                                          func=np.nanmean,
                                          cval=np.nan) #np.nanmedian(imgseq.as_array()))

    imgseq_reduced = neo.ImageSequence(images_reduced,
                                   units=imgseq.units,
                                   spatial_scale=imgseq.spatial_scale * macro_pixel_dim,
                                   macro_pixel_dim=macro_pixel_dim,
                                   sampling_rate=imgseq.sampling_rate,
                                   file_origin=imgseq.file_origin,
                                   t_start=imgseq.t_start)

    if 'array_annotations' in imgseq.annotations:
        del imgseq.annotations['array_annotations']

    imgseq_reduced.annotations.update(imgseq.annotations)

    if imgseq.name:
        imgseq_reduced.name = imgseq.name
    imgseq_reduced.annotations.update(macro_pixel_dim=macro_pixel_dim)
    imgseq_reduced.description = imgseq.description +  \
                "spatially downsampled ({}).".format(os.path.basename(__file__))

    return imgseq_reduced

def plot_downsampled_image(image, output_path):
    plt.figure()
    plt.imshow(image, interpolation='nearest', cmap='viridis', origin='lower')
    save_plot(output_path)
    return plt.gca()

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)

    imgseq_reduced = spatial_smoothing(imgseq, args.macro_pixel_dim)

    if args.output_img is not None:
        plot_downsampled_image(imgseq_reduced.as_array()[0], args.output_img)

    new_block = neo.Block()
    new_segment = neo.Segment()
    new_block.segments.append(new_segment)
    new_block.segments[0].imagesequences.append(imgseq_reduced)
    new_asig = imagesequence_to_analogsignal(imgseq_reduced)

    block.segments[0].analogsignals[0] = new_asig

    write_neo(args.output, block)



"""
Set the threshold between Up and Down states to a fixed value.
"""

import numpy as np
import argparse
from utils.io_utils import load_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output thresholds (numpy array)")
CLI.add_argument("--threshold", nargs='?', type=float, required=True,
                 help="")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    asig = load_neo(args.data, 'analogsignal')

    dim_t, channel_num = asig.shape

    np.save(args.output, np.ones(channel_num) * args.threshold)



import os
import neo
import matplotlib.pyplot as plt
import warnings
from snakemake.logging import logger
from pathlib import Path

def load_neo(filename, object='block', lazy=False, *args, **kwargs):
    try:
        filename = Path(filename)
        if filename.suffix == '.nix':
            kwargs.update(mode='ro')

        nio = neo.io.get_io(str(filename), *args, **kwargs)

        if lazy and nio.support_lazy:
            block = nio.read_block(lazy=lazy)
        # elif lazy and isinstance(io, neo.io.nixio.NixIO):
        #     with neo.NixIOFr(filename, *args, **kwargs) as nio:
        #         block = nio.read_block(lazy=lazy)
        else:
            block = nio.read_block()

    except Exception as e:
        # nio.close()
        raise e
    finally:
        if not lazy and hasattr(nio, 'close'):
            nio.close()

    if block is None:
        raise IOError(f'{filename} does not exist!')

    if object == 'block':
        return block
    elif object == 'analogsignal':
        return block.segments[0].analogsignals[0]
    else:
        raise IOError(f"{object} not recognized! Choose 'block' or 'analogsignal'.")


def write_neo(filename, block, *args, **kwargs):
    # muting saving imagesequences for now, since they do not yet
    # support array_annotations
    block.segments[0].imagesequences = []
    try:
        # for neo >= 0.12.0 filename can't contain '|'
        nio = neo.io.get_io(str(filename), *args, **kwargs)
        nio.write(block)
    except Exception as e:
        warnings.warn(str(e))
    finally:
        nio.close()
    return True


def save_plot(filename, dpi=300, **kwargs):
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    try:
        plt.savefig(fname=filename, dpi=dpi, bbox_inches='tight', **kwargs)
    except ValueError as ve:
        warnings.warn(str(ve))
        plt.subplots()
        plt.savefig(fname=filename)
    plt.close()
    return None



"""
Calculate the period between two consecutive waves for each wave.
"""

from pathlib import Path
import neo
import numpy as np
import quantities as pq
import matplotlib.pyplot as plt
import seaborn as sns
import os
import argparse
import scipy
import pandas as pd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name='wavefronts', objects="Event")[0]
    num_nonnan_channels = np.sum(np.isfinite(asig[0]).astype(int))

    wave_ids = np.sort(np.unique(evts.labels).astype(int))
    if wave_ids[0] == -1:
        wave_ids = np.delete(wave_ids, 0)

    wave_times = np.empty((2, asig.shape[1]), dtype=float)

    IWIs = np.empty((len(wave_ids),2))
    IWIs.fill(np.nan)

    t_unit = evts.times[0].dimensionality.string

    for wave_id in wave_ids[:-1]:
        wave_times.fill(np.nan)
        for i, wi in enumerate([wave_id, wave_id+1]):
            idx = np.where(evts.labels == str(wi))[0]
            trigger_channels = evts.array_annotations['channels'][idx]
            trigger_times = evts.times[idx]
            for channel, t in zip(trigger_channels, trigger_times):
                wave_times[i, int(channel)] = t
        inter_wave_intervals = wave_times[1] - wave_times[0]
        IWIs[wave_id, 0] = np.nanmean(inter_wave_intervals)
        IWIs[wave_id, 1] = np.nanstd(inter_wave_intervals)

    # transform to DataFrame
    df = pd.DataFrame(IWIs, columns=['inter_wave_interval', 'inter_wave_interval_std'])
    df['inter_wave_interval_unit'] = t_unit
    df[f'{args.event_name}_id'] = wave_ids
    df.to_csv(args.output)

    # ToDo
    fig, ax = plt.subplots()

    if np.isfinite(IWIs[:,0]).any():
        ax.hist(IWIs[:,0])
    
    ax.set_xlabel('inter-wave interval [s]')
    save_plot(args.output_img)



"""
Cluster similar waves into modes. 

Adapted from [Ruiz-Mejias et al. (2011)](https://doi.org/10.1523/JNEUROSCI.2517-15.2016)
"""

import argparse
import neo
import numpy as np
import pandas as pd
from warnings import warn
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from skimage.transform import resize
from scipy.interpolate import RBFInterpolator
from scipy.spatial.distance import cdist
import matplotlib as mpl
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_str, none_or_int
from utils.neo_utils import analogsignal_to_imagesequence, remove_annotations

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--min_trigger_fraction", "--MIN_TRIGGER_FRACTION",
                 nargs='?', type=float, default=.5,
                 help="minimum fraction of channels to be involved in a wave")
CLI.add_argument("--num_wave_neighbours", "--NUM_WAVE_NEIGHBOURS",
                 nargs='?', type=int, default=5,
                 help="number of similar waves to extrapolate nans from")
CLI.add_argument("--wave_outlier_quantile", "--WAVE_OUTLIER_QUANTILE",
                 nargs='?', type=float, default=.95,
                 help="percentage of similar waves to keep")
CLI.add_argument("--pca_dims", "--PCA_DIMS",
                 nargs='?', type=none_or_int, default=None,
                 help="reduce wave patterns to n dimensions before kmeans clustering")
CLI.add_argument("--num_kmeans_cluster", "--NUM_KMEANS_CLUSTER",
                 nargs='?', type=int, default=5,
                 help="number of wave modes to cluster with kmeans")
CLI.add_argument("--interpolation_step_size", "--INTERPOLATION_STEP_SIZE",
                 nargs='?', type=float, default=.2,
                 help="grid spacing for interpolation [0,1]")
CLI.add_argument("--interpolation_smoothing", "--INTERPOLATION_SMOOTHING",
                 nargs='?', type=float, default=0,
                 help="0: no smoothing, >0: more smoothing")

def build_timelag_dataframe(waves_evt, normalize=True):
    wave_ids = np.unique(waves_evt.labels).astype(int)
    channel_ids = np.unique(waves_evt.array_annotations['channels'])

    # init dataframe
    timelag_matrix = np.empty((len(wave_ids), len(channel_ids))) * np.nan
    timelag_df = pd.DataFrame(timelag_matrix,
                              index=wave_ids,
                              columns=channel_ids)
    timelag_df.index.name = 'wave_ids'
    timelag_df.columns.name = 'channel_ids'

    # fill timelag dataframe
    for i, trigger in enumerate(waves_evt):
        wave_id = int(trigger.labels)
        channel_id = waves_evt.array_annotations['channels'][i]
        # use only first trigger per channel and wave
        if np.isnan(timelag_df[channel_id][wave_id]):
            timelag_df[channel_id][wave_id] = trigger.magnitude

    if normalize:
        norm_func = lambda row: row - np.nanmean(row)
        timelag_df = timelag_df.apply(norm_func, axis=1)

    return timelag_df

def fill_nan_sites_from_similar_waves(timelag_df, num_neighbours=5,
                                      outlier_quantile=0.95):
    # ToDo: this doesn't work with too few waves!
    ## init arrays
    num_waves = timelag_df.index.size
    pair_indices = np.triu_indices(num_waves, 1)
    wavepair_distances = np.empty(len(pair_indices[0]), dtype=float) * np.nan
    neighbourhood_distance = np.empty(num_waves, dtype=float) * np.nan

    stds = np.array([])
    ## calculate wave distances
    for i, (a,b) in enumerate(zip(pair_indices[0], pair_indices[1])):
        wave_a, wave_b = timelag_df.iloc[a], timelag_df.iloc[b]
        wavepair_distances[i] = np.nanmean(np.abs(wave_a - wave_b))

    for row, wave_id in enumerate(timelag_df.index):
        ## sort other waves by their distance
        pair_pos = get_triu_indices_pos(i=row, N=num_waves)
        neighbour_distances = wavepair_distances[pair_pos]
        sort_idx = np.argsort(neighbour_distances)
        neighbour_distances = neighbour_distances[sort_idx]
        neighbourhood_distance[row] = np.mean(neighbour_distances[:num_neighbours])
        neighbour_pos = np.array([pair_indices[0][i] if pair_indices[0][i] != row
                                  else pair_indices[1][i] for i in pair_pos[sort_idx]])

        wave = timelag_df.loc[wave_id]
        if ~np.isnan(wave).any():
            continue
        ## average trigger times of channels of similar waves
        for channel in np.where(np.isnan(wave))[0]:
            similar_trigger = timelag_df.iloc[neighbour_pos, channel]
            similar_trigger = similar_trigger[~np.isnan(similar_trigger)]
            trigger_estimate = np.mean(similar_trigger[:num_neighbours])
            stds = np.append(stds, np.std(similar_trigger[:num_neighbours]))
            timelag_df.iloc[row, channel] = trigger_estimate

    # remove outlier waves in the quantile of wave distances
    if np.isnan(neighbourhood_distance).any():
        warn('Unexpected nan value in wave triggers!')
        return timelag_df
    
    if outlier_quantile < 1:
        q = np.quantile(neighbourhood_distance, outlier_quantile)
        keep_rows = np.where(neighbourhood_distance <= q)[0]
        timelag_df = timelag_df.iloc[keep_rows, :]
    return timelag_df

def get_triu_indices_pos(i, N):
    idx0_start = lambda x: x*N - np.sum(range(x+1))
    idx0 = np.array([k for k in range(idx0_start(i), idx0_start(i+1))], dtype=int)
    idx1 = np.array([i-1 + np.sum([(N-2-j) for j in range(k)])
                     for k in range(i)], dtype=int)
    return np.sort(np.append(idx0, idx1))

def pca_transform(timelag_matrix, dims=None):
    if dims is not None and dims > len(timelag_df):
        warn(f'Too few waves ({len(timelag_df)}) to perform a pca reduction '
           + f'to {dims} dims. Skipping.')
        dims = None
    if dims is None:
        return timelag_matrix.to_numpy()
    # n_samples x n_features
    if type(timelag_matrix) == pd.DataFrame:
        timelag_matrix = timelag_matrix.to_numpy()
    x_scaled = StandardScaler().fit_transform(timelag_matrix)
    pca_out = PCA(n_components=dims).fit(x_scaled)
    return pca_out.transform(x_scaled)

def kmeans_cluster_waves(timelag_matrix, n_cluster=7):
    n_waves = len(timelag_matrix)
    if n_cluster > n_waves:
        warn(f'Too few waves {n_waves} to determine {n_cluster} '
           + f'cluster. Reducing to {n_waves} cluster.')
        n_cluster = n_waves
    kmeans = KMeans(init="k-means++",
                    n_clusters=n_cluster,
                    tol=1e-10,
                    random_state=42,
                    algorithm='lloyd')

    return kmeans.fit(timelag_matrix)

def build_cluster_timelag_dataframe(timelag_df, cluster_labels):
    cluster_timelag_df = pd.DataFrame(index=np.unique(cluster_labels),
                                      columns=timelag_df.columns)
    cluster_timelag_df.index.name = 'cluster_ids'
    cluster_timelag_df.columns.name = 'channel_ids'

    for cluster in np.unique(cluster_labels):
        idx = np.where(cluster_labels == cluster)[0]
        avg_timelags = timelag_df.iloc[idx].mean(axis='index')
        cluster_timelag_df.loc[cluster] = avg_timelags
    return cluster_timelag_df

def arange_on_grid(df, channels, x_coords, y_coords):
    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1
    grid = np.empty((df.index.size, dim_y, dim_x)) * np.nan
    for row, wave in enumerate(df.iterrows()):
        wave = wave[1]
        for channel_id, timelag in zip(wave.index, wave):
            i = np.where(channel_id == channels)[0]
            x, y = x_coords[i], y_coords[i]
            grid[row,y,x] = timelag
    return grid

def wave_to_grid(wave_evt):
    timelag_df = build_timelag_dataframe(wave_evt)
    channels = np.unique(wave_evt.array_annotations['channels']).astype(int)
    annotation_idx = [np.argmax(wave_evt.array_annotations['channels']==channel) for channel in channels]
    x_coords = wave_evt.array_annotations['x_coords'][annotation_idx].astype(int)
    y_coords = wave_evt.array_annotations['y_coords'][annotation_idx].astype(int)
    grids = arange_on_grid(timelag_df, channels, x_coords, y_coords)
    return grids

def sample_wave_pattern(pattern_func, dim_x, dim_y, step):
    nx = round((dim_x-1)/step+1)
    ny = round((dim_y-1)/step+1)
    fy, fx = np.meshgrid(np.arange(0, ny*step, step),
                         np.arange(0, nx*step, step),
                         indexing='ij')
    fcoords = np.stack((fy,fx), axis=-1)
    fdim_y, fdim_x, _ = fcoords.shape
    fcoords = fcoords.reshape(-1,2)
    fcoords = fcoords if fdim_x > 1 else fcoords[:,0][:,None]
    wave_pattern = pattern_func(fcoords)
    return fx, fy, wave_pattern.reshape(fdim_y, fdim_x)

def interpolate_grid(grid, smoothing=0):
    y, x = np.where(np.isfinite(grid))
    coords = np.stack((y,x), axis=-1) if len(np.unique(x)) > 1 else y[:,None]
    rbf_func = RBFInterpolator(coords,
                               grid[y,x],
                               neighbors=None, smoothing=smoothing,
                               kernel='thin_plate_spline', epsilon=None,
                               degree=None)
    return rbf_func

def calc_cluster_distortions(feature_matrix, cluster_indices, cluster_centers):
    cluster_labels = np.unique(cluster_indices)
    cluster_dists = np.zeros(len(cluster_labels), dtype=float)

    for i, cluster_id in enumerate(cluster_labels):
        cluster_points = feature_matrix[np.where(cluster_indices==i)[0]]
        dists = cdist(cluster_points,
                      cluster_centers[i][np.newaxis,:],
                      metric='euclidean')
        cluster_dists[i] = np.sqrt(np.mean(dists**2))
    return cluster_dists

def plot_wave_modes(wavefronts_evt, wavemodes_evt):
    cmap = mpl.cm.get_cmap('coolwarm').copy()
    cmap.set_bad(color='white')
    n_modes = wavemodes_evt.annotations['n_modes']
    fig, axes = plt.subplots(ncols=n_modes+1,
                             figsize=(n_modes*4, 5),
                             gridspec_kw={'width_ratios':[1]*n_modes+[0.1]})

    int_step_size = wavemodes_evt.annotations['interpolation_step_size']

    for i, mode_id in enumerate(wavemodes_evt.annotations['mode_labels']):
        mode_count = wavemodes_evt.annotations['mode_counts'][i]
        mode_dist = wavemodes_evt.annotations['mode_distortions'][i]
        ax = axes[i]

        waves = wavefronts_evt[wavefronts_evt.array_annotations['wavemode'] == mode_id]
        waves_grid = wave_to_grid(waves)
        wavemode = wavemodes_evt[wavemodes_evt.labels.astype(int) == mode_id]
        mode_grid = wave_to_grid(wavemode)[0]

        vminmax = np.nanmax(mode_grid)
        img = ax.imshow(np.nanmean(waves_grid, axis=0), origin='lower',
                        interpolation='nearest',
                        cmap=cmap, alpha=0.5,
                        vmin=-vminmax, vmax=vminmax)

        dim_y, dim_x = mode_grid.shape
        if dim_x == 1:
            mode_grid = np.stack((np.squeeze(mode_grid), 
                                  np.squeeze(mode_grid)), axis=1)
        y, x = np.where(mode_grid)
        fx = x.reshape(mode_grid.shape) * int_step_size
        fy = y.reshape(mode_grid.shape) * int_step_size

        ctr = ax.contour(fx, fy, mode_grid, levels=9,
                         cmap=cmap, linewidths=2, alpha=1,
                         vmin=-vminmax, vmax=vminmax)
        for side in ['top','right','bottom','left']:
            ax.spines[side].set_visible(False)
        ax.tick_params(axis='both', which='both',
                       labelbottom=False, bottom=False, left=False)
        ax.set_yticklabels([])
        ax.set_xlabel(f'mode #{mode_id} | {mode_count} waves | var={mode_dist:.2f}')

    cbar = plt.colorbar(img, cax=axes[-1], ticks=[-vminmax/1.5, vminmax/1.5])
    cbar.ax.set_yticklabels(['wave start', 'wave end'],
                            rotation=90, va='center')
    return None

def clean_timelag_dataframe(df, min_trigger_fraction=.5, 
                            num_wave_neighbours=5, wave_outlier_quantile=1):
    # remove nan channels
    df.dropna(axis='columns', how='all', inplace=True)
 
    # remove small waves
    min_trigger_num = int(min_trigger_fraction * df.columns.size)
    df.dropna(axis='rows', thresh=min_trigger_num, inplace=True)

    # fill in nan sites with timelags from similar waves
    df = fill_nan_sites_from_similar_waves(df,
                                num_neighbours=num_wave_neighbours,
                                outlier_quantile=wave_outlier_quantile)
    return df


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    dim_t, num_channels = asig.shape

    ## BUILD TIMELAG MATRIX
    waves = block.filter(name='wavefronts', objects="Event")[0]
    waves = waves[waves.labels.astype(str) != '-1']

    if len(waves):
        timelag_df = build_timelag_dataframe(waves)

        ## CLEAN TIMELAG MATRIX
        timelag_df = clean_timelag_dataframe(timelag_df,
                                min_trigger_fraction=args.min_trigger_fraction,
                                num_wave_neighbours=args.num_wave_neighbours,
                                wave_outlier_quantile=args.wave_outlier_quantile)
    else:
        timelag_df = []

    if not len(timelag_df):
        warn("No waves found to cluster!")
        write_neo(args.output, block)
        if args.output_img is not None:
            save_plot(args.output_img)
        quit()

    ## CLUSTER WAVE MODES
    # PCA transform the timelag_matrix
    timelag_matrix_transformed = pca_transform(timelag_df, dims=args.pca_dims)

    # kmeans cluster the transformed timelag_matrix into modes
    kout = kmeans_cluster_waves(timelag_matrix_transformed,
                                n_cluster=args.num_kmeans_cluster)
    mode_ids = kout.labels_
    if len(mode_ids) != len(timelag_df):
        raise IndexError('Some waves are not assigned to a kmeans cluster!'
                         + f' {len(mode_ids)} != {len(timelag_df)}')

    mode_labels, mode_counts = np.unique(mode_ids, return_counts=True)

    mode_dists = calc_cluster_distortions(timelag_matrix_transformed,
                                          cluster_indices=mode_ids,
                                          cluster_centers=kout.cluster_centers_)

    # calculate the average timelags per mode
    mode_timelag_df = build_cluster_timelag_dataframe(timelag_df, mode_ids)

    # rearrange average mode timelags onto channel grid
    x_coords = asig.array_annotations['x_coords']
    y_coords = asig.array_annotations['y_coords']
    channels = np.arange(len(x_coords))
    mode_grids = arange_on_grid(mode_timelag_df, channels, x_coords, y_coords)
    n_modes, dim_y, dim_x = mode_grids.shape

    # interpolate average mode timelags as pattern on grid
    for i, cluster_grid in enumerate(mode_grids):
        pattern_func = interpolate_grid(cluster_grid, args.interpolation_smoothing)
        fx, fy, pattern = sample_wave_pattern(pattern_func,
                                              step=args.interpolation_step_size,
                                              dim_x=dim_x, dim_y=dim_y)
        interpolated_mode_grids = np.concatenate((interpolated_mode_grids,
                                                  pattern[np.newaxis,:])) \
                                  if i else pattern[np.newaxis,:]

    # add cluster labels as annotation to the wavefronts event
    evt_id, waves = [(i, evt) for i, evt in enumerate(block.segments[0].events) \
                                        if evt.name=='wavefronts'][0]

    mode_annotations = np.ones(waves.size, dtype=int) * (-1)
    for wave_id, mode_id in zip(timelag_df.index, mode_ids):
        index = np.where(waves.labels.astype(int) == wave_id)[0]
        mode_annotations[index] = mode_id

    waves.array_annotate(wavemode=mode_annotations)
    block.segments[0].events[evt_id] = waves

    # add clustered wave modes as additional event 'wavemodes'
    n_modes, inter_dim_y, inter_dim_x = interpolated_mode_grids.shape
    imgseq = analogsignal_to_imagesequence(asig)

    site_grid = np.isfinite(imgseq[0].as_array())
    interpolated_site_grid = resize(site_grid,
                                    output_shape=(inter_dim_y, inter_dim_x),
                                    mode='constant', cval=True,
                                    order=0)
    iys, ixs = np.where(interpolated_site_grid)
    n_sites = len(ixs)
    mode_trigger = np.empty(n_modes*n_sites)*np.nan

    for mode_id in range(n_modes):
        for site_id, (ix, iy) in enumerate(zip(ixs, iys)):
            mode_trigger[mode_id*n_sites + site_id] \
                                    = interpolated_mode_grids[mode_id, iy, ix]

    remove_annotations(waves)
    evt = neo.Event(mode_trigger * waves.units,
                    labels=np.repeat(mode_labels, n_sites).astype(str),
                    name='wavemodes',
                    mode_labels=mode_labels,
                    mode_counts=mode_counts[mode_labels],
                    mode_distortions=mode_dists,
                    interpolation_step_size=args.interpolation_step_size,
                    n_modes=n_modes,
                    pca_dims='None' if args.pca_dims is None else args.pca_dims,
                    **waves.annotations)
    evt.annotations['spatial_scale'] *= args.interpolation_step_size
    evt.array_annotations['x_coords'] = np.tile(ixs, n_modes).astype(int)
    evt.array_annotations['y_coords'] = np.tile(iys, n_modes).astype(int)
    evt.array_annotations['channels'] = np.tile(np.arange(n_sites), n_modes)

    block.segments[0].events.append(evt)
    # save output neo object

    write_neo(args.output, block)

    # plot and save output image
    plot_wave_modes(wavefronts_evt=block.segments[0].events[evt_id],
                    wavemodes_evt=evt)
    if args.output_img is not None:
        save_plot(args.output_img)



"""
Visualize the wave detection by means of clustering the detected trigger
in (time,x,y) space.
"""

import numpy as np
import quantities as pq
import argparse
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import seaborn as sns
import random
import warnings
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_float
from utils.neo_utils import time_slice

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str,
                 help="path of output file")
CLI.add_argument("--time_slice", nargs='?', type=none_or_float, default=None,
                 help="length of time_slice in seconds.")

def plot_clustering(events, ax=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
    N = len(np.unique(events.labels))
    if N:
        cmap = sns.husl_palette(N-1, h=.5, l=.6)
        cmap = random.sample([c for c in cmap], N-1)
        cmap = ListedColormap(['k']+cmap)

        ax.scatter(events.times,
                   events.array_annotations['x_coords'],
                   events.array_annotations['y_coords'],
                   c=[int(label) for label in events.labels],
                   cmap=cmap, s=2)
    else:
        warnings.warn('No trigger events to plot in clusters!')

    ax.set_xlabel('time [{}]'.format(events.times.dimensionality.string))
    ax.set_ylabel('x-pixel')
    ax.set_zlabel('y-pixel')
    ax.view_init(45, -75)
    return ax, cmap


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    evts = block.filter(name='wavefronts', objects="Event")[0]

    if len(evts):

        if args.time_slice is not None:
            asig = block.segments[0].analogsignals[0]
            t_stop = asig.t_start.rescale('s') + args.time_slice*pq.s
            evts = time_slice(evts, t_start=asig.t_start, t_stop=t_stop)

        ax, cmap = plot_clustering(evts)

    save_plot(args.output)



"""
Calculate the period between two consecutive waves for each wave and channel.
"""

import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--kernel", "--KERNEL", nargs='?', type=none_or_str, default=None,
                 help="derivative kernel")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

def calc_local_wave_intervals(evts):
    wave_labels = evts.labels.astype(int)
    unique_labels = np.sort(np.unique(wave_labels))
    unique_channels = np.sort(np.unique(evts.array_annotations['channels'].astype(int)))

    channel_idx_map = np.empty(np.max(unique_channels)+1) * np.nan
    for i, channel in enumerate(unique_channels):
        channel_idx_map[channel] = i

    trigger_collection = np.empty((len(unique_labels), len(unique_channels)),
                                  dtype=float) * np.nan
                          
    for (i, wave_id) in enumerate(unique_labels):
        wave_trigger_evts = evts[wave_labels == wave_id]

        channels = wave_trigger_evts.array_annotations['channels'].astype(int)
        
        channel_idx = channel_idx_map[channels].astype(int)
        trigger_collection[i, channel_idx] = wave_trigger_evts.times

    intervals = np.diff(trigger_collection, axis=0)
    intervals = intervals.reshape((len(unique_labels)-1)*len(unique_channels))

    mask = np.isfinite(intervals)
    intervals = intervals[mask]
    
    channel_ids = np.tile(unique_channels, len(unique_labels)-1)[mask]
    wave_ids = np.repeat(unique_labels[:-1], len(unique_channels))[mask]

    return wave_ids, channel_ids, intervals*evts.times.units



if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    imgseq = analogsignal_to_imagesequence(asig)

    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels != '-1']

    wave_ids, channel_ids, intervals = calc_local_wave_intervals(evts)

    # transform to DataFrame
    df = pd.DataFrame(intervals.magnitude, columns=['inter_wave_interval_local'])
    df['inter_wave_interval_local_unit'] = [intervals.dimensionality.string]*len(channel_ids)
    df['channel_id'] = channel_ids
    df[f'{args.event_name}_id'] = wave_ids

    df.to_csv(args.output)

    fig, ax = plt.subplots()
    ax.hist(1./intervals.magnitude[np.where(np.isfinite(1./intervals))[0]],
            bins=100, range=[0, 8])
    plt.xlabel('local rate of waves (Hz)', fontsize=7.)
    if args.output_img is not None:
        save_plot(args.output_img)



import numpy as np
import neo
import warnings
from itertools import product
import sys
from copy import copy
import quantities as pq
from pathlib import Path
from snakemake.logging import logger
from .parse import get_base_type, get_nan_value


def remove_annotations(objects, del_keys=['nix_name', 'neo_name']):
    if type(objects) != list:
        objects = [objects]
    for i, obj in enumerate(objects):
        for k in del_keys:
            if k in obj.annotations:
                del objects[i].annotations[k]
            if hasattr(obj, 'array_annotations') and k in obj.array_annotations:
                del objects[i].array_annotations[k]
    return None

def merge_analogsignals(asigs):
    # ToDo: to be replaced by neo utils functions
    if len(asigs) == 1:
        return asigs[0]

    min_length = np.min([len(asig.times) for asig in asigs])
    max_length = np.max([len(asig.times) for asig in asigs])
    if min_length != max_length:
        print('Warning: the length of the analog signals differs '\
            + 'between {} and {} '.format(min_length, max_length)\
            + 'All signals will be cut to the same length and merged '\
            + 'into one AnalogSignal object.')

    if len(np.unique([asig.sampling_rate for asig in asigs])) > 1:
        raise ValueError('The AnalogSignal objects have different '\
                       + 'sampling rates!')

    asig_array = np.empty((min_length, len(asigs)), dtype=float) * np.nan

    for channel_number, asig in enumerate(asigs):
        asig_array[:, channel_number] = np.squeeze(asig.as_array()[:min_length])

    merged_asig = neo.AnalogSignal(asig_array*asigs[0].units,
                                   sampling_rate=asigs[0].sampling_rate,
                                   t_start=asigs[0].t_start)

    for asig in asigs:
        for key, value in asig.array_annotations.items():
            if len(value) > 1:
                warnings.warn('Unexpected length of array annotations!')
                continue
            asig.annotations[key] = value[0]

    for key in asigs[0].annotations.keys():
        annotation_values = np.array([a.annotations[key] for a in asigs])
        try:
            if (annotation_values == annotation_values[0]).all():
                merged_asig.annotations[key] = annotation_values[0]
            else:
                merged_asig.array_annotations[key] = annotation_values
        except:
            warnings.warn(f'Can not merge annotation {key}!')
    return merged_asig


def flip_image(imgseq, axis=-1):
    # spatial axis 0 (~ 1) -> vertical
    # spatial axis 1 (~ 2) -> horizontal
    if len(imgseq.shape)==3 and axis==0:
        warnings.warn("Can not flip along time axis!"
                      "Interpreting axis=0 as first spatial axis (i.e. axis=1).")
        axis = 1

    flipped = np.flip(imgseq.as_array(), axis=axis)

    flipped_imgseq = imgseq.duplicate_with_new_data(flipped)

    if 'array_annotations' in imgseq.annotations.keys():
        flipped_imgseq.annotations['array_annotations'] = {}
        for key, value in imgseq.annotations['array_annotations'].items():
            flipped_imgseq.annotations['array_annotations'][key] \
                        = np.flip(value, axis=axis)
        flipped_imgseq.annotations['array_annotations']['y_coords'] \
                    = imgseq.annotations['array_annotations']['y_coords']

    return flipped_imgseq


def rotate_image(imgseq, rotation=0):
    # rotating clockwise
    if np.abs(rotation) <= 2*np.pi:
        # interpret as rad
        rotation = int(np.round(rotation/np.pi * 180, decimals=0))
    else:
        # interpret as deg
        pass

    nbr_of_rot90 = np.divide(rotation, 90)

    if np.mod(nbr_of_rot90, 1):
        nbr_of_rot90 = np.round(nbr_of_rot90, decimals=0)
        warnings.warn("Images can only be rotated in steps of 90 degrees. "
                     f"Rounding {rotation} deg to {nbr_of_rot90*90} deg.")

    rotated = np.rot90(imgseq.as_array(),
                       k=nbr_of_rot90,
                       axes=(-2,-1))

    rotated_imgseq = imgseq.duplicate_with_new_data(rotated)
    dim_t, dim_y, dim_x = rotated_imgseq.shape

    if 'array_annotations' in imgseq.annotations.keys():
        rotated_imgseq.annotations['array_annotations'] = {}
        for key, value in imgseq.annotations['array_annotations'].items():
            rotated_imgseq.annotations['array_annotations'][key] \
                    = np.rot90(value, k=nbr_of_rot90, axes=(-2,-1))

        y_coords, x_coords = np.meshgrid(range(dim_y),range(dim_x), indexing='ij')
        rotated_imgseq.annotations['array_annotations']['y_coords'] = y_coords
        rotated_imgseq.annotations['array_annotations']['x_coords'] = x_coords

    return rotated_imgseq

def robust_t(neo_obj, t_value=None, t_name='t_start', unit='s'):
    if t_value is None:
        if hasattr(neo_obj, t_name):
            t_value = getattr(neo_obj, t_name).rescale('s').magnitude
        else:
            raise Warning("t_start is not defined by the input or the object!")
    else:
        if isinstance(t_value, pq.Quantity):
            t_value = t_value.rescale('s').magnitude
        if hasattr(neo_obj, t_name):
            obj_t_start = neo_obj.t_start.rescale('s').magnitude
            obj_t_stop = neo_obj.t_stop.rescale('s').magnitude
            if not (obj_t_start <= t_value <= obj_t_stop):
                t_value = getattr(neo_obj, t_name).rescale('s').magnitude
    return t_value * pq.s.rescale(unit)


def time_slice(neo_obj, t_start=None, t_stop=None,
               lazy=False, channel_indexes=None, unit=pq.s):
    """
    Robustly time-slices neo.AnalogSignal, neo.IrregularSampledSignal,
    neo.ImageSequence, or neo.Event,
    with `t_start` and `t_stop` given in seconds.
    """
    if not lazy and not hasattr(neo_obj, 'time_slice'):
        raise TypeError(f"{neo_obj} has no function 'time_slice'!")
    if t_start is None and t_stop is None:
        return neo_obj

    t_start = robust_t(neo_obj, t_start, t_name='t_start')
    t_stop = robust_t(neo_obj, t_stop, t_name='t_stop')

    if lazy and hasattr(neo_obj, 'load'):
        return neo_obj.load(time_slice=(t_start, t_stop),
                            channel_indexes=channel_indexes)
    else:
        return neo_obj.time_slice(t_start, t_stop)


def imagesequence_to_analogsignal(imgseq):
    dim_t, dim_y, dim_x = imgseq.as_array().shape

    y_coords, x_coords = np.meshgrid(range(dim_y),range(dim_x), indexing='ij')
    x_coords = x_coords.reshape(dim_x * dim_y)
    y_coords = y_coords.reshape(dim_x * dim_y)

    imgseq_flat = imgseq.as_array().reshape((dim_t, dim_x * dim_y))

    asig = neo.AnalogSignal(signal=imgseq_flat,
                            units=imgseq.units,
                            dtype=imgseq.dtype,
                            t_start=imgseq.t_start,
                            sampling_rate=imgseq.sampling_rate,
                            file_origin=imgseq.file_origin,
                            description=imgseq.description,
                            name=imgseq.name,
                            array_annotations={'x_coords': x_coords,
                                               'y_coords': y_coords},
                            spatial_scale=imgseq.spatial_scale)

    remove_annotations(imgseq, del_keys=['nix_name', 'neo_name'])
    annotations = copy(imgseq.annotations)

    if 'array_annotations' in annotations.keys():
        try:
            for key, value in annotations['array_annotations'].items():
                asig.array_annotations[key] = value.reshape(dim_x*dim_y)

            if not (x_coords == asig.array_annotations['x_coords']).all() \
            or not (y_coords == asig.array_annotations['y_coords']).all():
                raise IndexError("Transformation of array_annotations for the " 
                               + "AnalogSignal went wrong!")
        except ValueError:
            warnings.warn("ImageSequence <-> AnalogSignal transformation "
                        + "changed the signal shape!")
        del annotations['array_annotations']

    asig.annotations.update(annotations)

    return asig


def analogsignal_to_imagesequence(asig):
    asig_array = asig.as_array()
    dim_t, dim_channels = asig_array.shape

    if 'x_coords' not in asig.array_annotations \
    or 'y_coords' not in asig.array_annotations:
        logger.error('AnalogSignal has no spatial information '
                   + 'as array_annotations "x_coords" "y_coords"!')

    x_coords = asig.array_annotations['x_coords'].astype(int)
    y_coords = asig.array_annotations['y_coords'].astype(int)

    if len(x_coords) != dim_channels or len(y_coords) != dim_channels:
        logger.error("Number of channels doesn't fit number of coordinates!")

    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1

    image_data = np.empty((dim_t, dim_y, dim_x), dtype=asig.dtype) * np.nan

    for channel, (x,y) in enumerate(zip(x_coords, y_coords)):
        image_data[:, y, x] = asig_array[:, channel]

    imgseq = neo.ImageSequence(image_data=image_data,
                               units=asig.units,
                               dtype=asig.dtype,
                               t_start=asig.t_start,
                               sampling_rate=asig.sampling_rate,
                               name=asig.name,
                               description=asig.description,
                               file_origin=asig.file_origin,
                               **asig.annotations)

    # add dict of array_annotations mapped to the 2D grid
    grid_array_annotations = {}
    for key, value in asig.array_annotations.items():
        dtype = get_base_type(value[0])
        nan_value = get_nan_value(dtype)

        grid_value = np.empty((dim_y, dim_x), dtype=value.dtype)
        grid_value.fill(nan_value)

        for channel, (x,y) in enumerate(zip(x_coords, y_coords)):
            grid_value[y, x] = value[channel]

        grid_array_annotations[key] = grid_value

    ys, xs = np.meshgrid(range(dim_y),range(dim_x), indexing='ij')
    # check if -1 in xy_coords
    if not (xs == grid_array_annotations['x_coords']).all() \
    or not (ys == grid_array_annotations['y_coords']).all():
        raise ValueError("Transformation of array_annotations for the "
                       + "imagesequence went wrong!")

    imgseq.annotate(array_annotations=grid_array_annotations)
    remove_annotations(imgseq, del_keys=['nix_name', 'neo_name'])
    return imgseq


def add_empty_sites_to_analogsignal(asig):
    x_coords = asig.array_annotations['x_coords']
    y_coords = asig.array_annotations['y_coords']
    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1
    yx_coords = list(zip(y_coords, x_coords))

    asig_array = asig.as_array()
    dim_t, dim_channels = asig_array.shape
    num_grid_channels = dim_x * dim_y
    num_nan_channels = num_grid_channels - dim_channels
    
    if num_nan_channels == 0:
        return asig

    nan_signals = np.empty((dim_t, num_nan_channels)) * np.nan
    new_asig_array = np.append(asig_array, nan_signals, axis=1)

    grid_coords = list(product(range(dim_y), range(dim_x)))
    nan_coords = list(set(grid_coords).difference(yx_coords))
    y_nan_coords = np.array(nan_coords)[:,0]
    x_nan_coords = np.array(nan_coords)[:,1]
    num_nans = len(nan_coords)

    new_asig = asig.duplicate_with_new_data(new_asig_array)

    # add nans into array_annotations for empty sites
    for key, values in asig.array_annotations.items():
        nan_value = np.array([get_nan_value(get_base_type(values))])
        new_values = np.append(values, np.repeat(nan_value, num_nans))
        if type(values) == pq.Quantity:
            new_values = new_values.magnitude * values.units
        new_asig.array_annotations[key] = new_values

    new_asig.array_annotate(x_coords=np.append(x_coords, x_nan_coords), 
                            y_coords=np.append(y_coords, y_nan_coords))

    return new_asig



"""
Detect trigger times (i.e., state transition / local wavefronts onsets) 
by finding crossing of a set phase-value in the channel signals.
"""

import argparse
import neo
import numpy as np
import quantities as pq
from scipy.signal import hilbert
from scipy.stats import zscore
import matplotlib.pyplot as plt
import seaborn as sns
from utils.io_utils import load_neo, write_neo, save_plot
from utils.neo_utils import time_slice, remove_annotations
from utils.parse import none_or_int, none_or_float
from pathlib import Path

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--img_dir", nargs='?', type=Path,
                 default=None, help="path of figure directory")
CLI.add_argument("--img_name", nargs='?', type=str,
                 default='hilbert_phase_channel0.png',
                 help='example image filename for channel 0')
CLI.add_argument("--transition_phase", nargs='?', type=float, default=-1.570796,
                 help="phase to use as threshold for the upward transition")
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")
CLI.add_argument("--plot_tstart", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--plot_tstop",  nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")

def detect_transitions(asig, transition_phase):
    # ToDo: replace with elephant function
    signal = asig.as_array()
    dim_t, channel_num = signal.shape

    hilbert_signal = hilbert(signal, axis=0)
    hilbert_phase = np.angle(hilbert_signal)

    def _detect_phase_crossings(phase):
        # detect phase crossings from below phase to above phase
        is_larger = hilbert_phase > phase
        positive_crossings = ~is_larger & np.roll(is_larger, -1, axis=0)
        positive_crossings = positive_crossings[:-1]

        # select phases within [-pi, pi]
        real_crossings = np.real(hilbert_signal[:-1]) > np.imag(hilbert_signal[:-1])
        crossings = real_crossings & positive_crossings

        # arrange transitions times per channel
        times = asig.times[:-1]
        crossings_list = [times[crossings[:,channel]].magnitude
                          for channel in range(channel_num)]
        return crossings_list

    # UP transitions: A change of the hilbert phase from < transtion_phase
    #                 to > transition_phase, followed by a peak (phase = 0).

    peaks = _detect_phase_crossings(0)
    transitions = _detect_phase_crossings(transition_phase)

    up_transitions = np.array([])
    channels = np.array([], dtype=int)

    for channel_id, (channel_peaks, channel_transitions) in enumerate(zip(peaks, transitions)):
        channel_up_transitions = np.array([])
        if channel_peaks is not None:
            for peak in channel_peaks:
                distance_to_peak = peak - np.array(channel_transitions)
                distance_to_peak = distance_to_peak[distance_to_peak > 0]
                if distance_to_peak.size:
                    trans_idx = np.argmin(distance_to_peak)
                    channel_up_transitions = np.append(channel_up_transitions,
                                                       channel_transitions[trans_idx])
        channel_up_transitions = np.unique(channel_up_transitions)
        up_transitions = np.append(up_transitions, channel_up_transitions)
        channels = np.append(channels,
                             np.ones_like(channel_up_transitions, dtype=int)*channel_id)

    # save transitions as Event labels:'UP', array_annotations: channels
    sort_idx = np.argsort(up_transitions)

    evt = neo.Event(times=up_transitions[sort_idx]*asig.times.units,
                    labels=['UP'] * len(up_transitions),
                    name='transitions',
                    array_annotations={'channels':channels[sort_idx]},
                    trigger_detection='hilbert_phase',
                    hilbert_transition_phase=transition_phase,
                    description='transitions from down to up states. '\
                               +'annotated with the channel id ("channels").')

    for key in asig.array_annotations.keys():
        evt_ann = {key : asig.array_annotations[key][channels[sort_idx]]}
        evt.array_annotations.update(evt_ann)

    remove_annotations(asig, del_keys=['nix_name', 'neo_name'])
    evt.annotations.update(asig.annotations)
    return evt


def plot_hilbert_phase(asig, event, channel):
    signal = asig.as_array()[:,channel]

    hilbert_signal = hilbert(signal, axis=0)
    hilbert_phase = np.angle(hilbert_signal)

    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()

    ax.plot(asig.times.rescale('s'), zscore(signal), label='signal')
    ax.plot(asig.times.rescale('s'), hilbert_phase, label='hilbert phase')

    for t, c in zip(event.times, event.array_annotations['channels']):
        if c == channel:
            ax.axvline(t.rescale('s'), color='k')

    ax.set_title('Channel {}'.format(channel))
    ax.set_xlabel('time [{}]'.format(asig.times.units.dimensionality.string))
    ax.legend()
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]

    args.plot_tstart = asig.t_start if args.plot_tstart is None else args.plot_tstart
    args.plot_tstop = asig.t_stop if args.plot_tstop is None else args.plot_tstop

    transition_event = detect_transitions(asig, args.transition_phase)

    block.segments[0].events.append(transition_event)

    write_neo(args.output, block)

    if args.plot_channels[0] is not None:
        for channel in args.plot_channels:
            plot_hilbert_phase(asig=time_slice(asig, args.plot_tstart, args.plot_tstop),
                               event=time_slice(transition_event, args.plot_tstart, args.plot_tstop),
                               channel=int(channel))
            output_path = args.img_dir / args.img_name.replace('_channel0', f'_channel{channel}')
            save_plot(output_path)



"""
Extract the annotations of Neo objects and structure them in a DataFrame
to complement a wave characterization.
"""

import argparse
import numpy as np
import pandas as pd
import quantities as pq
import re
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str
from utils.neo_utils import remove_annotations

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")
CLI.add_argument("--ignore_keys", "--IGNORE_KEYS", nargs='*', type=str, default=[],
                 help="neo.Event annotations keys to not include in dataframe")
CLI.add_argument("--include_keys", "--INCLUDE_KEYS", nargs='*', type=str, default=[],
                 help="neo object annotations keys to include in dataframe")
CLI.add_argument("--profile", "--PROFILE", nargs='?', type=none_or_str, default=None,
                 help="profile name")
                    
def add_annotations_to_df(df, annotations, include_keys=[]):
    use_all_keys = not bool(len(include_keys))

    for key, value in annotations.items():
        key_is_relevant = use_all_keys or key in include_keys

        if key_is_relevant and key not in df.columns:
            if type(value) == pq.Quantity:
                df[f'{key}_unit'] = value.dimensionality.string
                value = value.magnitude
            df[key] = value

    return df

def get_corresponding_array_values(a, b):
    '''
    a -> b
    '''
    a_labels = np.unique(a)
    mapping = {}
    for i, a_label in enumerate(a_labels):
        idx = np.where(a_label == a)[0]
        b_label_values = b[idx]
        if (b_label_values == b_label_values[0]).all():
            mapping[a_label] = b_label_values[0]
        else:
            return False
    return mapping

def add_array_annotations_to_df(df, array_annotations, labels, index_name,
                                include_keys=[]):
    use_all_keys = not bool(len(include_keys))

    for key, value in array_annotations.items():
        key_is_relevant = use_all_keys or key in include_keys

        if key_is_relevant and key not in df.columns:
            mapping = get_corresponding_array_values(labels, value)
            if not mapping: continue

            annotation_df = pd.DataFrame(mapping.items(),
                                         columns=[index_name, key])
            df = df.merge(annotation_df, how='outer', on=index_name)
    return df

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()
    args.ignore_keys = [re.sub('[\[\],\s]', '', key) for key in args.ignore_keys]
    args.include_keys = [re.sub('[\[\],\s]', '', key) for key in args.include_keys]
    if len(args.include_keys):
        args.ignore_keys = []

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels.astype(str) != '-1']

    remove_annotations(evts, del_keys=['nix_name', 'neo_name']+args.ignore_keys)
    remove_annotations(asig, del_keys=['nix_name', 'neo_name']+args.ignore_keys)

    ids = np.sort(np.unique(evts.labels).astype(int))
    df = pd.DataFrame(ids, columns=[f'{args.event_name}_id'])

    for annotations in [evts.annotations, asig.annotations]:
        df = add_annotations_to_df(df, annotations, args.include_keys)

    df = add_array_annotations_to_df(df, evts.array_annotations,
                                     labels=evts.labels.astype(int),
                                     index_name=f'{args.event_name}_id',
                                     include_keys=args.include_keys)

    df['profile'] = [args.profile] * len(df.index)
    df['sampling_rate'] = asig.sampling_rate.magnitude
    df['sampling_rate_unit'] = asig.sampling_rate.dimensionality.string
    df['recording_length'] = (asig.t_stop - asig.t_start).magnitude
    df['recording_length_unit'] = asig.t_start.dimensionality.string
    df['dim_x'] = int(max(asig.array_annotations['x_coords']))+1
    df['dim_y'] = int(max(asig.array_annotations['y_coords']))+1

    df.to_csv(args.output)

    # ToDo
    save_plot(args.output_img)



"""
Subtract the background of the input data by subtracting the mean of each channel.
"""

import numpy as np
import matplotlib.pyplot as plt
import argparse
from pathlib import Path
import os
from utils.io_utils import load_neo, write_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str,
                 help="path of output image", default=None)
CLI.add_argument("--output_array", nargs='?', type=none_or_str,
                 help="path of output numpy array", default=None)

def shape_frame(value_array, xy_coords):
    dim_x = np.max(xy_coords[:,0]) + 1
    dim_y = np.max(xy_coords[:,1]) + 1
    frame = np.empty((dim_y, dim_x)) * np.nan
    for pixel, (x,y) in zip(value_array, xy_coords):
        frame[int(y), int(x)] = pixel
    return frame

def plot_frame(frame):
    fig, ax = plt.subplots()
    ax.imshow(frame, interpolation='nearest', cmap=plt.cm.gray, origin='lower')
    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    return ax

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]
    signal = asig.as_array()
    background = np.nanmean(signal, axis=0)
    signal -= background

    if args.output_img or args.output_array is not None:
        xy_coords = np.array([(x,y) for x,y in
                           zip(asig.array_annotations['x_coords'],
                               asig.array_annotations['y_coords'])],
                           dtype=int)
        frame = shape_frame(background, xy_coords)
        if args.output_array is not None:
            np.save(args.output_array, frame)
        if args.output_img is not None:
            plot_frame(frame)
            save_plot(args.output_img)

    new_asig = asig.duplicate_with_new_data(signal)
    new_asig.array_annotations = asig.array_annotations
    new_asig.description += "The mean of each channel was subtracted ({})."\
                        .format(os.path.basename(__file__))
    block.segments[0].analogsignals[0] = new_asig

    write_neo(args.output, block)



"""
Detect trigger times (i.e., state transition / local wavefronts onsets) 
by applying a threshold to each channel signal.
"""

import neo
import numpy as np
import argparse
from utils.io_utils import load_neo, write_neo
from utils.neo_utils import remove_annotations

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--thresholds", nargs='?', type=str, required=True,
                 help="path of thresholds (numpy array)")

def threshold(asig, threshold_array):
    dim_t, channel_num = asig.shape
    th_signal = asig.as_array()\
              - np.repeat(threshold_array[np.newaxis, :], dim_t, axis=0)
    state_array = th_signal > 0
    rolled_state_array = np.roll(state_array, 1, axis=0)

    all_times = np.array([])
    all_channels = np.array([], dtype=int)
    all_labels = np.array([])
    for label, func in zip(['UP',        'DOWN'],
                           [lambda x: x, lambda x: np.bitwise_not(x)]):
        trans = np.where(func(np.bitwise_not(rolled_state_array))\
                       * func(state_array))
        channels = trans[1]
        times = asig.times[trans[0]]
        
        if not len(times):
            raise ValueError("The chosen threshold lies not within the range "\
                           + "of the signal values!")

        all_channels = np.append(all_channels, channels)
        all_times = np.append(all_times, times)
        all_labels = np.append(all_labels, np.array([label for _ in times]))

    sort_idx = np.argsort(all_times)

    evt = neo.Event(times=all_times[sort_idx]*asig.times.units,
                    labels=all_labels[sort_idx],
                    name='transitions',
                    trigger_detection='threshold',
                    array_annotations={'channels':all_channels[sort_idx]},
                    threshold=threshold_array,
                    description='Transitions between down and up states with '\
                               +'labels "UP" and "DOWN". '\
                               +'Annotated with the channel id ("channels").')

    for key in asig.array_annotations.keys():
        evt_ann = {key : asig.array_annotations[key][all_channels[sort_idx]]}
        evt.array_annotations.update(evt_ann)

    remove_annotations(asig, del_keys=['nix_name', 'neo_name'])
    evt.annotations.update(asig.annotations)
    return evt


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]

    transition_event = threshold(asig, np.load(args.thresholds))

    block.segments[0].events.append(transition_event)

    write_neo(args.output, block)



import numpy as np
import itertools
import warnings
from scipy.ndimage import convolve as conv
from types import SimpleNamespace

simple_3x3 = SimpleNamespace(
             x=np.array([[-0, 0, 0],
                         [-1, 0, 1],
                         [-0, 0, 0]], dtype=float),
             y=np.array([[-0, -1, -0],
                         [ 0,  0,  0],
                         [ 0,  1,  0]], dtype=float))

prewitt_3x3 = SimpleNamespace(
             x=np.array([[-1, 0, 1],
                         [-1, 0, 1],
                         [-1, 0, 1]], dtype=float),
             y=np.array([[-1, -1, -1],
                         [ 0,  0,  0],
                         [ 1,  1,  1]], dtype=float))

scharr_3x3 = SimpleNamespace(
             x=np.array([[ -3, 0,  3],
                         [-10, 0, 10],
                         [ -3, 0,  3]], dtype=float),
             y=np.array([[-3, -10, -3],
                         [ 0,   0,  0],
                         [ 3,  10,  3]], dtype=float))

sobel_3x3 = SimpleNamespace(
            x=np.array([[-1, 0, 1],
                        [-2, 0, 2],
                        [-1, 0, 1]], dtype=float),
            y=np.array([[-1, -2, -1],
                        [ 0,  0,  0],
                        [ 1,  2,  1]], dtype=float))

sobel_5x5 = SimpleNamespace(
            x=np.array([[ -5, -4, 0,  4,  5],
                        [ -8,-10, 0, 10,  8],
                        [-10,-20, 0, 20, 10],
                        [ -8,-10, 0, 10,  8],
                        [ -5, -4, 0,  4,  5]], dtype=float),
            y=np.array([[ -5, -8, -10, -8, -5],
                        [ -4,-10, -20,-10,  4],
                        [  0,  0,   0,  0,  0],
                        [  4, 10,  20, 10,  4],
                        [  5,  8,  10,  8,  5]], dtype=float))

sobel_7x7 = SimpleNamespace(
            x=np.array([[-3/18, -2/13, -1/10, 0,  1/10, 2/13, 3/18],
                        [-3/13, -2/8 , -1/5 , 0,  1/5 , 2/8 , 3/13],
                        [-3/10, -2/5 , -1/2 , 0,  1/2 , 2/5 , 3/10],
                        [-3/9 , -2/4 , -1/1 , 0,  1/1 , 2/4 , 3/9 ],
                        [-3/10, -2/5 , -1/2 , 0,  1/2 , 2/5 , 3/10],
                        [-3/13, -2/8 , -1/5 , 0,  1/5 , 2/8 , 3/13],
                        [-3/18, -2/13, -1/10, 0,  1/10, 2/13, 3/18]], dtype=float),
            y=np.array([[-3/18, -3/13, -3/10, -3/9,  -3/10, -3/13, -3/18],
                        [-2/13, -2/8 , -2/5 , -2/4,  -2/5 , -2/8 , -2/13],
                        [-1/10, -1/5 , -1/2 , -1/1,  -1/2 , -1/5 , -1/10],
                        [    0,    0,     0 ,    0,     0 ,    0 ,     0],
                        [ 1/10,  1/5 ,  1/2 ,  1/1,   1/2 ,  1/5 ,  1/10],
                        [ 2/13,  2/8 ,  2/5 ,  2/4,   2/5 ,  2/8 ,  2/13],
                        [ 3/18,  3/13,  3/10,  3/9,   3/10,  3/13,  3/18]], dtype=float))

for kernel in [simple_3x3, prewitt_3x3, scharr_3x3,
               sobel_3x3, sobel_5x5, sobel_7x7]:
    kernel.x /= np.sum(np.abs(kernel.x))
    kernel.y /= np.sum(np.abs(kernel.y))

def get_kernel(kernel_name):
    if kernel_name.lower() in ['simple_3x3', 'simple']:
        return simple_3x3
    elif kernel_name.lower() in ['prewitt_3x3', 'prewitt']:
        return prewitt_3x3
    elif kernel_name.lower() in ['scharr_3x3', 'scharr']:
        return scharr_3x3
    elif kernel_name.lower() in ['sobel_3x3', 'sobel']:
        return sobel_3x3
    elif kernel_name.lower() in ['sobel_5x5']:
        return sobel_5x5
    elif kernel_name.lower() in ['sobel_7x7']:
        return sobel_7x7
    else:
        warnings.warn(f'Deriviative name {kernel_name} is not implemented, '
                     + 'using sobel filter instead.')
        return sobel_3x3


def nan_conv2d(frame, kernel, kernel_center=None):
    dx, dy = kernel.shape
    dimx, dimy = frame.shape
    dframe = np.empty((dimx, dimy))*np.nan

    if kernel_center is None:
        kernel_center = [int((dim-1)/2) for dim in kernel.shape]

    # inverse kernel to mimic behavior or regular convolution algorithm
    k = kernel[::-1, ::-1]
    ci = dx - 1 - kernel_center[0]
    cj = dy - 1 - kernel_center[1]

    # loop over each frame site
    for i,j in zip(*np.where(np.isfinite(frame))):
        site = frame[i,j]

        # loop over kernel window for frame site
        window = np.zeros((dx,dy), dtype=float)*np.nan
        for di,dj in itertools.product(range(dx), range(dy)):

            # kernelsite != 0, framesite within borders and != nan
            if k[di,dj] and 0 <= i+di-ci < dimx and 0 <= j+dj-cj < dimy \
                        and np.isfinite(frame[i+di-ci,j+dj-cj]):
                sign = -1*np.sign(k[di,dj])
                window[di,dj] = sign * (site - frame[i+di-ci,j+dj-cj])

        xi, yi = np.where(np.logical_not(np.isnan(window)))
        if np.sum(np.logical_not(np.isnan(window))) > dx*dy/10:
            dframe[i,j] = np.average(window[xi,yi], weights=abs(k[xi,yi]))
    return dframe

norm_angle = lambda p: -np.mod(p + np.pi, 2*np.pi) + np.pi

def phase_conv2d(frame, kernel, kernel_center=None):
    dx, dy = kernel.shape
    dimx, dimy = frame.shape
    dframe = np.zeros_like(frame)

    if kernel_center is None:
        kernel_center = [int((dim-1)/2) for dim in kernel.shape]

    # inverse kernel to mimic behavior or regular convolution algorithm
    k = kernel[::-1, ::-1]
    ci = dx - 1 - kernel_center[0]
    cj = dy - 1 - kernel_center[1]

    # loop over kernel window for each frame site
    for i,j in zip(*np.where(np.isfinite(frame))):
        phase = frame[i,j]
        dphase = np.zeros((dx,dy), dtype=float)

        for di,dj in itertools.product(range(dx), range(dy)):

            # kernelsite != 0, framesite within borders and != nan
            if k[di,dj] and 0 <= i+di-ci < dimx and 0 <= j+dj-cj < dimy \
            and np.isfinite(frame[i+di-ci,j+dj-cj]):
                sign = -1*np.sign(k[di,dj])
                # pos = clockwise from phase to frame[..]
                dphase[di,dj] = sign*norm_angle(phase-frame[i+di-ci,j+dj-cj])

        if dphase.any():
            dframe[i,j] = np.average(dphase, weights=abs(k)) / np.pi
    return dframe



"""
Calculate the number of triggers involved in each wave.
"""

import numpy as np
import argparse
import pandas as pd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name=args.event_name, objects="Event")[0]

    wave_ids = np.sort(np.unique(evts.labels).astype(int))

    number_of_triggers = np.empty(len(wave_ids), dtype=float)

    for i, wave_id in enumerate(wave_ids):
        idx = np.where(evts.labels == str(wave_id))[0]
        number_of_triggers[i] = len(evts.times[idx])

    # transform to DataFrame
    df = pd.DataFrame(number_of_triggers, columns=['number_of_triggers'])
    df[f'{args.event_name}_id'] = wave_ids
    df.to_csv(args.output)

    # ToDo
    save_plot(args.output_img)



"""
Calculate the planarity each waves.
"""

import argparse
import os
import numpy as np
import neo
import pandas as pd
import matplotlib.pyplot as plt
import copy
import seaborn as sns
from utils.io_utils import load_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--alignment_threshold", nargs='?', type=float, default=.85,
                 help="threshold for alignment of velocity vectors at transitions")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")
                    
def label_planar(waves_event, vector_field, times, threshold):
    labels = np.unique(waves_event.labels)
    planarity = np.zeros(len(labels), dtype=float)

    for i, label in enumerate(labels):
        idx = np.where(label == waves_event.labels)[0]

        t_idx = [np.argmax(times >= t) for t
                 in waves_event.times[idx]]
        x = waves_event.array_annotations['x_coords'][idx]
        y = waves_event.array_annotations['y_coords'][idx]

        wave_directions = vector_field[t_idx, y.astype(int), x.astype(int)]
        norm = np.array([np.linalg.norm(w) for w in wave_directions])
        wave_directions /= norm

        planarity[i] = np.linalg.norm(np.mean(wave_directions))

    is_planar = planarity > threshold

    df = pd.DataFrame(planarity, columns=['planarity'])
    df['is_planar'] = is_planar
    return df


def plot_planarity(waves_event, vector_field, times, wave_id, skip_step=1, ax=None):
    label = np.unique(waves_event.labels)[wave_id]

    idx = np.where(label == waves_event.labels)[0]

    t_idx = np.array([np.argmax(times >= t) for t
                      in waves_event.times[idx]])
    x = waves_event.array_annotations['x_coords'][idx]
    y = waves_event.array_annotations['y_coords'][idx]

    wave_directions = vector_field[t_idx, y.astype(int), x.astype(int)]
    norm = np.array([np.linalg.norm(w) for w in wave_directions])
    wave_directions /= norm

    palette = sns.husl_palette(len(np.unique(t_idx))+1, h=0.3, l=0.4)[:-1]

    if ax is None:
        fig, ax = plt.subplots()

    area = copy.copy(np.real(vector_field[0]))
    area[np.where(np.isfinite(area))] = -.7
    ax.imshow(area, interpolation='nearest', origin='lower',
              vmin=-1, vmax=1, cmap='Greys')

    for i, frame_t in enumerate(np.unique(t_idx)):
        frame_i = np.where(frame_t == t_idx)[0].astype(int)
        xi = x[frame_i]
        yi = y[frame_i]
        ti = t_idx[frame_i]
        frame = vector_field[ti, yi.astype(int), xi.astype(int)].magnitude
        ax.quiver(xi, yi, np.real(frame), np.imag(frame),
                  # units='width', scale=max(frame.shape)/(10*skip_step),
                  # width=0.15/max(frame.shape),
                  color=palette[i], alpha=0.8,
                  label='{:.3f} s'.format(times[frame_t].rescale('s').magnitude))

    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylabel(f'pixel size {vector_field.spatial_scale:.2f}')
    start_t = np.min(waves_event.times[idx]).rescale('s').magnitude
    stop_t = np.max(waves_event.times[idx]).rescale('s').magnitude
    ax.set_xlabel('{:.3f} - {:.3f} s'.format(start_t, stop_t))
    ax.set_title('planarity {:.3f}'.format(np.linalg.norm(np.mean(wave_directions))))
    plt.legend(bbox_to_anchor=(1,1), loc='upper left')
    return ax


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    vec_asig = block.filter(name='optical_flow', objects="AnalogSignal")[0]
    optical_flow = analogsignal_to_imagesequence(vec_asig)
    
    wavefront_evt = block.filter(name=args.event_name, objects="Event")[0]
    wavefront_evt = wavefront_evt[wavefront_evt.labels.astype('str') != '-1']


    planar_labels = label_planar(waves_event=wavefront_evt,
                                 vector_field=optical_flow,
                                 times=vec_asig.times,
                                 threshold=args.alignment_threshold)
    planar_labels[f'{args.event_name}_id'] = np.unique(wavefront_evt.labels)
    planar_labels.to_csv(args.output)

    dim_t, dim_y, dim_x = optical_flow.shape
    skip_step = int(min([dim_x, dim_y]) / 50) + 1

    for i, wave_id in enumerate(np.unique(wavefront_evt.labels)):
        fig, ax = plt.subplots()
        plot_planarity(waves_event=wavefront_evt,
                       vector_field=optical_flow,
                       times=vec_asig.times,
                       skip_step=skip_step,
                       wave_id=i,
                       ax=ax)
        save_plot(os.path.join(os.path.dirname(args.output),
                               f'wave_{wave_id}.png'))
        if not i:
            save_plot(args.output_img)
        plt.close()



"""
Detect and characterize critical points in the optical flow vector field at each
time point.
"""

import argparse
import numpy as np
import neo
import matplotlib.pyplot as plt
from shapely.geometry import LineString
from utils.io_utils import load_neo, write_neo, save_plot
from utils.neo_utils import analogsignal_to_imagesequence

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
                    
def detect_critical_points(imgseq, times):
    frames = imgseq.as_array()
    if frames.dtype != np.complex128:
        raise ValueError("Vector field values must be complex numbers!")

    dim_t, dim_x, dim_y = frames.shape

    frame_ids = np.array([], dtype=int)
    labels = np.array([], dtype=str)
    x = np.array([], dtype=float)
    y = np.array([], dtype=float)
    trace = np.array([], dtype=float)
    det = np.array([], dtype=float)
    extend = np.array([], dtype=int)
    winding_number = np.array([], dtype=int)
    # winding_number = np.array([], dtype=float) # ToDo

    for i in range(dim_t):
        # ToDo: parallelize
        Y, X = np.meshgrid(np.arange(dim_y), np.arange(dim_x), indexing='ij')
        ZR = np.real(frames[i])
        ZI = np.imag(frames[i])
        contourR = plt.contour(X, Y, ZR, levels=[0])
        contourI = plt.contour(X, Y, ZI, levels=[0])

        for xy in get_line_intersections(contourR, contourI):
            x = np.append(x, xy[0])
            y = np.append(y, xy[1])
            frame_ids = np.append(frame_ids, i)

            J = jacobian(xy, ZR, ZI)
            trace_i = np.trace(J)
            det_i = np.linalg.det(J)
            labels = np.append(labels, classify_critical_point(det=det_i,
                                                               trace=trace_i))
            trace = np.append(trace, trace_i)
            det = np.append(det, det_i)

            extend_i, winding_number_i = calc_winding_number(xy, frames[i])
            extend = np.append(extend, extend_i)
            winding_number = np.append(winding_number, winding_number_i)

    evt = neo.Event(name='critical_points',
                    times=times[frame_ids],
                    labels=labels)
    evt.array_annotations.update({'x':x, 'y':y, 
                                  'trace':trace, 'det':det,
                                  'extend':extend, 
                                  'winding_number':winding_number})
    return evt


def jacobian(xy, fA, fB):
    # dA/dy  dA/dx
    # dB/dy  dB/dx
    x, y = int(np.round(xy[0])), int(np.round(xy[1]))
    J = np.zeros((2,2))
    dim_y, dim_x = fA.shape
    if y+1 < dim_y:
        J[0,0] = fA[y+1, x] - fA[y, x]
        J[1,0] = fB[y+1, x] - fB[y, x]
    else:
        J[0,0] = fA[y, x] - fA[y-1, x]
        J[1,0] = fB[y, x] - fB[y-1, x]
    if x+1 < dim_x:
        J[0,1] = fA[y, x+1] - fA[y, x]
        J[1,1] = fB[y, x+1] - fB[y, x]
    else:
        J[0,1] = fA[y, x] - fA[y, x-1]
        J[1,1] = fB[y, x] - fB[y, x-1]
    return J


def classify_critical_point(det, trace):
    if det > 0:
        if trace**2 > 4*det:
            if trace > 0 :
                return 'node stable'
            else:
                return 'node unstable'
        else:
            if trace > 0:
                return 'focus stable'
            else:
                return 'focus unstable'
    else:
        return 'saddle'


def calc_winding_number(xy, frame):
    px, py = int(np.round(xy[0])), int(np.round(xy[1]))
    dim_y, dim_x = frame.shape
    max_r = np.ceil(min([dim_x, dim_y])/2)
    prev_winding_number = 0
    for r in range(1, int(max_r)):
        # is circle intersecting with the image border?
        if px + r >= dim_x or px - r <= 0:
            break
        elif py + r >= dim_y or py - r <= 0:
            break
        else:
            Y, X = np.indices((dim_y, dim_x))
            circle = np.abs(np.hypot(py-Y,px-X) - r) < 0.5

        # extracting counter-clock-wise indices of circle with radius r
        ccw = np.array([], dtype=[('x',int), ('y',int)])
        # x0     x1     y0    y1     sx   sy
        # 0  0   xp xp  0 py  py -1  + -  - -
        # xp xp  -1 -1  0 py  py -1  + -  + +
        for x0, x1, y0, y1 in zip([0 , px, px, 0], # Quadrants TL, BL, BR, TR
                                  [px, -1, -1, px],
                                  [0 ,  0, py, py],
                                  [py, py, -1, -1]):
            circle_yi, circle_xi = np.where(circle[y0:y1, x0:x1])
            circle_xi += x0
            circle_yi += y0
            circle_idx = np.array([(-np.sign(x1)*yi, np.sign(y1)*xi)
                                    for xi,yi in zip(circle_xi, circle_yi)],
                                  dtype=[('x', int), ('y', int)])
            ccw = np.append(ccw, np.sort(circle_idx, order=['y', 'x']))

        # sum differences of subsequent vector angles around circle
        circle_values = frame[np.abs(ccw['y']), np.abs(ccw['x'])]
        circle_angles = np.angle(circle_values) + np.pi
        circle_angles = np.append(circle_angles, circle_angles[0])

        # ToDo: solve more pythonic
        sum_angles = 0
        for i, phi in enumerate(circle_angles[:-1]):
            dphi = circle_angles[i+1] - phi
            if dphi > np.pi:
                dphi -= 2*np.pi
            elif dphi < -np.pi:
                dphi += 2*np.pi
            sum_angles += dphi
        winding_number = sum_angles/(2*np.pi)

        winding_number = np.round(winding_number)
        if r != 1 and prev_winding_number != winding_number:
            break
        else:
            prev_winding_number = winding_number
    return r-1, prev_winding_number


def get_line_intersections(contourA, contourB):
    points = []
    for path in contourA.collections[0].get_paths():
        if len(path.vertices) > 1:
            lineA = LineString(path.vertices)
        else:
            lineA = None

        for path in contourB.collections[0].get_paths():
            if len(path.vertices) > 1 and lineA is not None:
                lineB = LineString(path.vertices)

                intersection = lineA.intersection(lineB)

                if not intersection.is_empty:
                    if intersection.type == 'MultiPoint':
                        points += [(i.x, i.y) for i in intersection]
                    elif intersection.type == 'Point':
                        points += [(intersection.x, intersection.y)]
    return points


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()
    block = load_neo(args.data)

    asig = block.filter(name='optical_flow', objects="AnalogSignal")[0]
    imgseq = analogsignal_to_imagesequence(asig)

    crit_point_evt = detect_critical_points(imgseq,
                                    block.segments[0].analogsignals[0].times)

    block.segments[0].events.append(crit_point_evt)

    write_neo(args.output, block)



#!/usr/bin/env python
# encoding: utf8
"""
Collaborative Brain Wave Analysis Pipeline (Cobrawap)
"""

import argparse
import inspect
import logging
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from pprint import pformat

sys.path.append(str(Path(inspect.getfile(lambda: None)).parent))
sys.path.append(str(Path(inspect.getfile(lambda: None)).parent / "pipeline"))
from cmd_utils import (
    create_new_configfile,
    get_available_blocks,
    get_config,
    get_initial_available_stages,
    get_profile,
    get_setting,
    input_block,
    input_profile,
    input_stage,
    is_profile_name_valid,
    load_config_file,
    locate_str_in_list,
    print_settings,
    read_stage_output,
    set_setting,
    setup_entry_stage,
    working_directory,
)

log = logging.getLogger()
logging.basicConfig(level=logging.INFO)


try:
    STAGES = get_setting("stages")
except Exception as e:
    log.debug(e)
    try:
        STAGES = get_initial_available_stages()
    except Exception as e:
        log.debug(e)
        STAGES = {}

CLI = argparse.ArgumentParser(prog="cobrawap")


def get_parser():
    return CLI


# Utility arguments
CLI.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="print additional logging information",
)
CLI.add_argument("-V", "--version", action="version")
CLI.set_defaults(command=None)

# Initialization
subparsers = CLI.add_subparsers(help="")
CLI_init = subparsers.add_parser(
    "init",
    help="initialize the cobrawap directories (required only once)",
)
CLI_init.set_defaults(command="init")
CLI_init.add_argument(
    "--output_path",
    type=Path,
    default=None,
    help="directory where the analysis output is stored "
    "[default: '~/cobrawap_output/']",
)
CLI_init.add_argument(
    "--config_path",
    type=Path,
    default=None,
    help="directory where the analysis config files are "
    "stored [default: '~/cobrawap_config/']",
)

# Show Settings
CLI_settings = subparsers.add_parser(
    "settings",
    help="display the content of ~/.cobrawap/config",
)
CLI_settings.set_defaults(command="settings")


# Configuration
CLI_create = subparsers.add_parser(
    "create",
    help="create configuration for a new dataset",
)
CLI_create.set_defaults(command="create")
CLI_create.add_argument(
    "--data_path",
    type=Path,
    nargs="?",
    default=None,
    help="full path to the dataset to be analyzed (or where it will be stored)",
)
CLI_create.add_argument(
    "--loading_script_name",
    nargs="?",
    type=Path,
    default=None,
    help="name of the data specific loading script "
    "(in <config_path>/stage01_data_entry/scripts/)",
)
CLI_create.add_argument(
    "--profile",
    type=str,
    nargs="?",
    default=None,
    help="profile name of this dataset/application "
    "(see profile name conventions in documentation)",
)
CLI_create.add_argument(
    "--parent_profile",
    type=str,
    nargs="?",
    default=None,
    help="optional parent profile name "
    "(see profile name conventions in documentation)",
)

# Additional configurations
CLI_profile = subparsers.add_parser(
    "add_profile",
    help="create a new configuration for an existing dataset",
)
CLI_profile.set_defaults(command="add_profile")
CLI_profile.add_argument(
    "--profile",
    type=str,
    nargs="?",
    default=None,
    help="profile name of this dataset/application "
    "(see profile name conventions in documentation)",
)
CLI_profile.add_argument(
    "--stages",
    type=str,
    nargs="*",
    choices=list(STAGES.keys()),
    default=None,
    help="selection of pipeline stages to configure",
)
CLI_profile.add_argument(
    "--parent_profile",
    type=str,
    nargs="?",
    default=None,
    help="optional parent profile name from which to "
    "initialize the new config "
    "[default: basic template]",
)

# Run
CLI_run = subparsers.add_parser(
    "run",
    help="run the analysis pipeline on the selected "
    "input and with the specified configurations",
)
CLI_run.set_defaults(command="run")
CLI_run.add_argument(
    "--profile",
    type=str,
    nargs="?",
    default=None,
    help="name of the config profile to be analyzed",
)

# Stage
CLI_stage = subparsers.add_parser(
    "run_stage",
    help="execute an individual stage",
)
CLI_stage.set_defaults(command="run_stage")
CLI_stage.add_argument(
    "--stage",
    type=str,
    nargs="?",
    default=None,
    choices=[s for pair in zip(STAGES.keys(), STAGES.values()) for s in pair],
    help="select individual stage to execute",
)
CLI_stage.add_argument(
    "--profile",
    type=str,
    nargs="?",
    default=None,
    help="name of the config profile to be analyzed",
)

# Block
CLI_block = subparsers.add_parser(
    "run_block",
    help="execute an individual block method on some input",
)
CLI_block.set_defaults(command="run_block")
CLI_block.add_argument(
    "--stage",
    type=str,
    nargs="?",
    default=None,
    choices=[s for pair in zip(STAGES.keys(), STAGES.values()) for s in pair],
    help="select individual stage to execute",
)
CLI_block.add_argument(
    "--block",
    type=str,
    nargs="?",
    default=None,
    help="select individual block to execute",
)
CLI_block.add_argument(
    "--block_help",
    action="store_true",
    help="display the help text of the block script",
)


def main():
    "Start main CLI entry point."
    args, unknown = CLI.parse_known_args()
    # log.info("this is a regular print statement")
    # log.debug("this is a verbose print statement")

    if args.verbose:
        log.setLevel(logging.DEBUG)
    log.debug(pformat(args))

    if args.command == "init":
        log.info("initializing Cobrawap")
        initialize(**vars(args))

    elif args.command == "settings":
        log.info("display settings at ~/.cobrawap/config")
        print_settings(**vars(args))

    elif args.command == "create":
        log.info("creating a set of config files")
        create(**vars(args))

    elif args.command == "add_profile":
        log.info("creating a new config files")
        add_profile(**vars(args))

    elif args.command == "run":
        log.info("executing Cobrawap")
        run(**vars(args), extra_args=unknown)

    elif args.command == "run_stage":
        log.info("executing Cobrawap stage")
        run_stage(**vars(args), extra_args=unknown)

    elif args.command == "run_block":
        log.info("executing Cobrawap block")
        run_block(**vars(args), block_args=unknown)

    elif args.command is None:
        CLI.print_help(sys.stderr)
        CLI.parse_args()

    else:
        log.info(f"{args.command} not known!")

    return None


def initialize(output_path=None, config_path=None, **kwargs):
    # set output_path
    if output_path is None:
        output_path = (
            Path(
                input("Output directory " "[default: ~/cobrawap_output]: ")
                or Path("~") / "cobrawap_output"
            )
            .expanduser()
            .resolve()
        )
    output_path.mkdir(exist_ok=True)
    if not output_path.is_dir():
        raise ValueError(f"{output_path} is not a valid directory!")

    set_setting(dict(output_path=str(output_path)))

    # set config_path
    if config_path is None:
        config_path = (
            Path(
                input("Config directory " "[default: ~/cobrawap_configs]: ")
                or Path("~") / "cobrawap_configs"
            )
            .expanduser()
            .resolve()
        )
    config_path.mkdir(parents=True, exist_ok=True)
    if not config_path.is_dir():
        raise ValueError(f"{config_path} is not a valid directory!")

    set_setting(dict(config_path=str(config_path)))

    # set pipeline path
    pipeline_path = Path(__file__).parents[1] / "cobrawap" / "pipeline"
    set_setting(dict(pipeline_path=str(pipeline_path.resolve())))

    # set available stages
    set_setting(dict(stages=get_initial_available_stages()))
    stages = get_setting("stages")

    # populate config_path with template config files
    if any(config_path.iterdir()):
        overwrite = (
            input(
                f"The config directory {config_path} already exists "
                "and is not empty. Create template config files "
                "anyway? [y/N] "
            ).lower()
            == "y"
            or False
        )
    else:
        overwrite = True
    if overwrite:
        for stage_number, stage in stages.items():
            stage_config_path = config_path / stage / "configs"
            stage_config_path.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                pipeline_path / stage / "configs" / "config_template.yaml",
                stage_config_path / "config.yaml",
            )

        pipeline_config_path = config_path / "configs"
        pipeline_config_path.mkdir(parents=True, exist_ok=True)
        shutil.copy(
            pipeline_path / "configs" / "config_template.yaml",
            pipeline_config_path / "config.yaml",
        )

        stage01_script_path = config_path / stages["1"] / "scripts"
        stage01_script_path.mkdir(parents=True, exist_ok=True)
        shutil.copy(
            pipeline_path / stages["1"] / "scripts" / "enter_data_template.py",
            stage01_script_path / "enter_data_template.py",
        )

    return None


def create(
    profile=None,
    parent_profile=None,
    data_path=None,
    loading_script_name=None,
    **kwargs,
):
    profile, parent_profile = get_profile(
        profile=profile, parent_profile=parent_profile
    )
    base_name = parent_profile if parent_profile else profile

    for stage_number, stage in get_setting("stages").items():
        config_name = profile if "1" in str(stage_number) else base_name
        create_new_configfile(
            stage=stage,
            profile=config_name,
            parent=parent_profile,
        )

    setup_entry_stage(
        profile=profile,
        parent_profile=parent_profile,
        data_path=data_path,
        loading_script_name=loading_script_name,
    )
    return None


def add_profile(
    profile=None,
    parent_profile=None,
    stages=None,
    data_path=None,
    loading_script_name=None,
    **kwargs,
):
    profile, parent_profile = get_profile(
        profile=profile, parent_profile=parent_profile
    )
    # get stage selection
    stages = ""
    while not stages:
        stages = input(
            "To which stages should this profile be applied? "
            f"{list(get_setting('stages').keys())}: "
        )
        try:
            stages = stages.replace("'", "")
            stages = re.split(",|\s", stages)
            stages = [stage for stage in stages if stage]
        except Exception as e:
            log.info(e)
            stages = ""

    for stage_number in stages:
        create_new_configfile(
            stage_number=stage_number, profile=profile, parent=parent_profile
        )

    if any("1" in stage for stage in stages):
        setup_entry_stage(
            profile=profile,
            parent_profile=parent_profile,
            data_path=data_path,
            loading_script_name=loading_script_name,
        )
    return None


def run(profile=None, extra_args=None, **kwargs):
    # # allow for positional profile argument
    # if profile is None and extra_args and extra_args[0][0] != "-":
    #     profile = extra_args.pop(0)

    # select profile
    profile = input_profile(profile=profile)

    # set runtime config
    pipeline_path = Path(get_setting("pipeline_path"))

    # execute snakemake
    snakemake_args = ["snakemake", "-c1", "--config", f"PROFILE={profile}"]
    log.info(f'Executing `{" ".join(snakemake_args+extra_args)}`')

    with working_directory(pipeline_path):
        subprocess.run(snakemake_args + extra_args)

    return None


def run_stage(stage=None, profile=None, extra_args=None, **kwargs):
    # # allow for positional stage argument
    # if stage is None and extra_args and extra_args[0][0] != "-":
    #     stage = extra_args.pop(0)

    stage = input_stage(stage=stage)
    profile = input_profile(profile=profile)

    # get settings
    pipeline_path = Path(get_setting("pipeline_path"))
    config_path = Path(get_setting("config_path"))
    output_path = Path(get_setting("output_path"))

    # lookup stage input file
    pipeline_config_path = config_path / "configs" / "config.yaml"
    config_dict = load_config_file(pipeline_config_path)
    stage_idx = locate_str_in_list(config_dict["STAGES"], stage)
    # stage_idx_global = locate_str_in_list([v for k,v in stages.items()], stage)
    if stage_idx is None:
        raise IndexError(
            "Make sure that the selected stage is also specified "
            "in your top-level config in the list `STAGES`!"
        )

    stage_config_path = get_config(
        config_dir=config_path / stage,
        config_name=f"config_{profile}.yaml",
        get_path_instead=True,
    )

    if stage_idx > 0:
        prev_stage = config_dict["STAGES"][stage_idx - 1]
        prev_stage_config_path = get_config(
            config_dir=config_path / prev_stage,
            config_name=f"config_{profile}.yaml",
            get_path_instead=True,
        )
        prev_config_name = Path(prev_stage_config_path).name
        prev_output_name = read_stage_output(
            stage=prev_stage,
            config_dir=config_path,
            config_name=prev_config_name,
        )
        stage_input = output_path / profile / prev_stage / prev_output_name
        extra_args = [f"STAGE_INPUT={stage_input}"] + extra_args

    # descend into stage folder
    pipeline_path = pipeline_path / stage

    # append stage specific arguments
    extra_args = extra_args + ["--configfile", f"{stage_config_path}"]

    # execute snakemake
    snakemake_args = ["snakemake", "-c1", "--config", f"PROFILE={profile}"]
    log.info(f'Executing `{" ".join(snakemake_args+extra_args)}`')

    with working_directory(pipeline_path):
        subprocess.run(snakemake_args + extra_args)

    return None


def run_block(stage=None, block=None, block_args=None, block_help=False, **kwargs):
    # # allow for positional block argument
    # if block is None and block_args and block_args[0][0] != "-":
    #     block = block_args.pop(0)

    stage = input_stage(stage=stage)
    block = input_block(stage=stage, block=block)

    block_dir = Path(get_setting("pipeline_path")) / stage / "scripts"

    if block_help:
        block_args += ["--help"]

    # execute block
    myenv = os.environ.copy()
    myenv["PYTHONPATH"] = ":".join(sys.path)
    subprocess.run(["python", str(block_dir / f"{block}.py")] + block_args, env=myenv)
    return None


if __name__ == "__main__":
    main()



"""
Check whether the input data representation adheres to the stage's requirements.

Additionally prints a short summary of the data attributes.
"""

import numpy as np
import argparse
from pathlib import Path
from utils.io_utils import load_neo
from utils.neo_utils import analogsignal_to_imagesequence, imagesequence_to_analogsignal

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if len(block.segments) > 1:
        print("More than one Segment found; all except the first one " \
            + "will be ignored.")
    if len(block.segments[0].analogsignals) > 1:
        print("More than one AnalogSignal found; all except the first one " \
            + "will be ignored.")

    asig = block.segments[0].analogsignals[0]

    imgseq = analogsignal_to_imagesequence(asig)
    asig2 = imagesequence_to_analogsignal(imgseq)

    if asig.shape != asig2.shape:
        raise ValueError("AnalogSignal doesn't include empty grid sites. "
                      + f"Reshape {asig.shape} to {asig2.shape} according to "
                         "x/y_coords. You may use `add_empty_sites_to_analogsignal` "
                         "from the utils.neo_utils module.")

    print('Recording Time:\t\t', asig.t_stop - asig.t_start)
    print('Sampling Rate:\t\t', asig.sampling_rate)
    print('Spatial Scale:\t\t', asig.annotations['spatial_scale'])

    num_channels = np.count_nonzero(~np.isnan(np.sum(asig, axis=0)))
    print('Number of Channels:\t', num_channels)

    x_coords = asig.array_annotations['x_coords']
    y_coords = asig.array_annotations['y_coords']

    dim_x, dim_y = np.max(x_coords)+1, np.max(y_coords)+1

    print('Grid Dimensions:\t', f'{dim_x} x {dim_y}')
    print('Empty Grid Sites:\t', dim_x*dim_y - num_channels)



"""
Filter between a highpass and a lowpass frequency using a butterworth filter.
"""

import argparse
import quantities as pq
import os
from elephant.signal_processing import butter
from utils.io_utils import load_neo, write_neo
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--highpass_frequency", nargs='?', type=none_or_float,
                 default=None, help="lower bound of frequency band in Hz")
CLI.add_argument("--lowpass_frequency", nargs='?', type=none_or_float,
                 default=None, help="upper bound of frequency band in Hz")
CLI.add_argument("--order", nargs='?', type=int, default=2,
                 help="order of the filter function")
CLI.add_argument("--filter_function", nargs='?', type=str, default='filtfilt',
                 help="filter function used in the scipy backend")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = butter(block.segments[0].analogsignals[0],
                  highpass_frequency=args.highpass_frequency*pq.Hz,
                  lowpass_frequency=args.lowpass_frequency*pq.Hz,
                  order=args.order,
                  filter_function=args.filter_function)

    asig.array_annotations = block.segments[0].analogsignals[0].array_annotations
    asig.annotate(highpass_frequency=args.highpass_frequency*pq.Hz,
                  lowpass_frequency=args.lowpass_frequency*pq.Hz,
                  filter_order=args.order)

    asig.description += "Frequency filtered with [{}, {}]Hz order {} "\
                        .format(args.highpass_frequency,
                                args.lowpass_frequency,
                                args.order)\
                      + " using {} scipy algorithm.({}). "\
                        .format(args.filter_function,
                                os.path.basename(__file__))
    block.segments[0].analogsignals[0] = asig

    write_neo(args.output, block)



"""
Z-score the signal of each channel.
"""

import argparse
from pathlib import Path
from elephant.signal_processing import zscore
from utils.io_utils import load_neo, write_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output",  nargs='?', type=Path, required=True,
                 help="path of output file")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    zscore(block.segments[0].analogsignals[0], inplace=True)

    write_neo(args.output, block)



"""
Create a plot of the channel-wise and average power spectrum density.
"""

import numpy as np
import quantities as pq
import argparse
import matplotlib.pyplot as plt
import seaborn as sns
from elephant.spectral import welch_psd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output figure")
CLI.add_argument("--highpass_frequency", nargs='?', type=none_or_float,
                 default='None', help="lower bound of frequency band in Hz")
CLI.add_argument("--lowpass_frequency", nargs='?', type=none_or_float,
                 default='None', help="upper bound of frequency band in Hz")
CLI.add_argument("--psd_frequency_resolution", nargs='?', type=float, default=5,
                 help="frequency resolution of the power spectrum in Hz")
CLI.add_argument("--psd_overlap", nargs='?', type=float, default=0.5,
                 help="overlap parameter for Welch's algorithm [0-1]")

def plot_psd(frequencies, psd, highpass_frequency, lowpass_frequency):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()
    for channel in psd:
        ax.semilogy(frequencies, channel, alpha=0.7)
    ax.semilogy(frequencies, np.mean(psd, axis=0), linewidth=2,
                color='k', label='channel average')
    ax.set_title('Power Spectrum')
    ax.set_xlabel('frequency [Hz]')
    ax.set_ylabel('power spectral density')

    if highpass_frequency is not None and lowpass_frequency is not None:
        left = highpass_frequency if highpass_frequency is not None else 0
        right = lowpass_frequency if lowpass_frequency is not None else ax.get_xlim()[1]
        ax.axvspan(left, right, alpha=0.2, color='k', label='filtered region')

    plt.legend()
    return fig


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    asig = load_neo(args.data, 'analogsignal')

    freqs, psd = welch_psd(asig,
                           frequency_resolution=args.psd_frequency_resolution*pq.Hz,
                           window='hann',
                           overlap=args.psd_overlap)

    plot_psd(frequencies=freqs,
             psd=psd,
             highpass_frequency=args.highpass_frequency,
             lowpass_frequency=args.lowpass_frequency)

    save_plot(args.output)



"""
Calculate the timing of each wave.
"""

import numpy as np
import matplotlib.pyplot as plt
import argparse
import pandas as pd
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image")
CLI.add_argument("--time_point", "--TIME_STAMP_POINT", nargs='?', type=str, default='start',
                 help="when to register the time for a wave [start, middle, end]")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if args.time_point == 'start':
        time_stamp_func = np.min
    elif args.time_point == 'end':
        time_stamp_func = np.max
    elif args.time_point == 'middle':
        time_stamp_func = np.mean
    else:
        raise InputError('')

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels.astype('str') != '-1']

    wave_ids = np.sort(np.unique(evts.labels).astype(int))

    time_stamps = np.empty(len(wave_ids), dtype=float)

    t_unit = evts.times[0].dimensionality.string

    for i, wave_id in enumerate(wave_ids):
        idx = np.where(evts.labels == str(wave_id))[0]
        time_stamps[i] = time_stamp_func(evts.times[idx])


    fig, ax = plt.subplots(figsize=(15,2))
    for i, wave_id in enumerate(wave_ids):
        idx = np.where(evts.labels == str(wave_id))[0]
        t0, t1 = np.min(evts.times[idx]), np.max(evts.times[idx])
        ax.plot([t0,t1], [1,1], marker='|', color='b')
    ax.set_ylim((0,2))
    ax.set_xlabel(f'time [{t_unit}]')
    ax.set_title('wave occurences')
    save_plot(args.output_img)

    # transform to DataFrame
    df = pd.DataFrame(time_stamps, columns=['time_stamp'])
    df['time_stamp_unit'] = t_unit
    df[f'{args.event_name}_id'] = wave_ids

    df.to_csv(args.output)



"""
Plot an excerpt of the input data and corresponding trigger times.
"""

import matplotlib.pyplot as plt
import seaborn as sns
import argparse
import os
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_int, none_or_float
from utils.neo_utils import time_slice

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str,
                 required=True, help="path of output directory")
CLI.add_argument("--filename", nargs='?', type=str,
                 default='trigger_times_channel0.png',
                 help='example filename for channel 0')
CLI.add_argument("--plot_tstart", nargs='?', type=none_or_float, default=0,
                 help="start time in seconds")
CLI.add_argument("--plot_tstop", nargs='?', type=none_or_float, default=10,
                 help="stop time in seconds")
CLI.add_argument("--plot_channels", nargs='+', type=none_or_int, default=None,
                 help="list of channels to plot")
                    
def plot_trigger_times(asig, event, channel):
    sns.set(style='ticks', palette="deep", context="notebook")
    fig, ax = plt.subplots()

    ax.plot(asig.times, asig.as_array()[:,channel], label='signal')

    times = [time for i, time in enumerate(event.times)
             if event.array_annotations['channels'][i]==channel]
    labels = [label for i, label in enumerate(event.labels)
             if event.array_annotations['channels'][i]==channel]

    if 'DOWN'.encode('UTF-8') in labels or 'DOWN' in labels:
        # plot up states
        plot_states(times, labels, ax,
                    t_start=asig.t_start, t_stop=asig.t_stop, label='UP states')
    elif 'UP'.encode('UTF-8') in labels or 'UP' in labels:
        # plot only up transitions
        for i, trans_time in enumerate(times):
            ax.axvline(trans_time, c='k',
                       label='UP transitions' if not i else '')
    else:
        print("Warning: No 'UP' (or 'DOWN') transition events "\
            + f"in channel {channel} found!")

    ax.set_title(f'Channel {channel}')
    ax.set_xlabel(f'time [{asig.times.units.dimensionality.string}]')
    ax.set_ylabel(f'signal [{asig.units.dimensionality.string}]')

    plt.legend()
    return fig


def plot_states(times, labels, ax, t_start, t_stop, label=''):
    if labels[0] == 'DOWN'.encode('UTF-8') or labels[0] == 'DOWN':
        ax.axvspan(t_start, times[0], alpha=0.5, color='red')
    if labels[-1] == 'UP'.encode('UTF-8') or labels[-1] == 'UP':
        ax.axvspan(times[-1], t_stop, alpha=0.5, color='red')

    for i, (time, label) in enumerate(zip(times, labels)):
        if (label == 'UP'.encode('UTF-8') or label == 'UP') \
            and i < len(times)-1:
            ax.axvspan(time, times[i+1], alpha=0.5, color='red',
                       label='UP' if not i else '')
    return None


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)
    asig = block.segments[0].analogsignals[0]

    args.plot_tstart = asig.t_start if args.plot_tstart is None else args.plot_tstart
    args.plot_tstop = asig.t_stop if args.plot_tstop is None else args.plot_tstop
    # slice signals
    asig = time_slice(asig, args.plot_tstart, args.plot_tstop)

    # get transition events
    event = block.filter(name='transitions', objects="Event")[0]
    event = event.time_slice(args.plot_tstart, args.plot_tstop)

    for channel in args.plot_channels:
        plot_trigger_times(asig=asig,
                           event=event,
                           channel=channel)
        output_path = os.path.join(args.output,
                                   args.filename.replace('_channel0', f'_channel{channel}'))
        save_plot(output_path)



"""
Remove detected triggers (state transitions) when the corresponding Up and Down 
states are shorter than a minimum duration.
"""

import numpy as np
import neo
import argparse
import quantities as pq
from utils.io_utils import load_neo, write_neo
from utils.parse import str_to_bool

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path to output data in neo format")
CLI.add_argument("--min_up_duration", nargs='?', type=float, default=0.005,
                 help="minimum duration of UP states in seconds")
CLI.add_argument("--min_down_duration", nargs='?', type=float, default=0.005,
                 help="minimum duration of DOWN states in seconds")
CLI.add_argument("--remove_down_first", nargs='?', type=str_to_bool, default=True,
                 help="If True, remove short down states first")


def remove_short_states(evt, min_duration, start_label='UP', stop_label='DOWN'):
    # assumes event times to be sorted
    del_idx = np.array([], dtype=int)

    for channel in np.unique(evt.array_annotations['channels']):
        # select channel
        c_idx = np.where(channel == evt.array_annotations['channels'])[0]
        c_times = evt.times[c_idx]
        c_labels = evt.labels[c_idx]

        # sepearate start and stop times
        start_idx = np.where(start_label == c_labels)[0]
        stop_idx = np.where(stop_label == c_labels)[0]
        start_times = c_times[start_idx]
        stop_times = c_times[stop_idx]

        # clean borders
        leading_stops = np.argmax(stop_times > start_times[0])
        stop_idx = stop_idx[leading_stops:]
        stop_times = stop_times[leading_stops:]
        start_times = start_times[:len(stop_times)]

        # find short states
        short_state_idx = np.where((stop_times-start_times).rescale('s')
                                    < min_duration.rescale('s'))[0]

        # remove end points of short states
        del_idx = np.append(del_idx, c_idx[stop_idx[short_state_idx]])
        if not start_label == stop_label:
            # remove start points of short states
            del_idx = np.append(del_idx, c_idx[start_idx[short_state_idx]])

    cleaned_evt = neo.Event(times=np.delete(evt.times.rescale('s'), del_idx)*pq.s,
                            labels=np.delete(evt.labels, del_idx),
                            name=evt.name,
                            description=evt.description)
    cleaned_evt.annotations = evt.annotations
    for key in evt.array_annotations:
        cleaned_evt.array_annotations[key] = np.delete(evt.array_annotations[key],
                                                       del_idx)
    return cleaned_evt


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    evt_idx, evt = [(i,ev) for i, ev in enumerate(block.segments[0].events)
                    if ev.name == 'transitions'][0]


    if 'DOWN' in evt.labels:
        if args.remove_down_first:
            evt = remove_short_states(evt, args.min_down_duration*pq.s,
                                      start_label='DOWN', stop_label='UP')
            evt = remove_short_states(evt, args.min_up_duration*pq.s,
                                      start_label='UP', stop_label='DOWN')
        else:
            evt = remove_short_states(evt, args.min_up_duration*pq.s,
                                      start_label='UP', stop_label='DOWN')
            evt = remove_short_states(evt, args.min_down_duration*pq.s,
                                      start_label='DOWN', stop_label='UP')
    else:
        remove_short_states(evt, (args.min_down_duration+args.min_up_duration)*pq.s,
                            start_label='UP', stop_label='UP')

    evt.annotations.update(min_up_duration=args.min_up_duration*pq.s)
    evt.annotations.update(min_down_duration=args.min_down_duration*pq.s)

    block.segments[0].events[evt_idx] = evt

    write_neo(args.output, block)



"""
Divide the signal in each channel by their max/mean/median value.
"""

import numpy as np
import argparse
import neo
import quantities as pq
import os
import sys
from utils.io_utils import write_neo, load_neo

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--normalize_by", nargs='?', type=str, default='mean',
                 help="division factor: 'max', 'mean', or 'median'")

def normalize(asig, normalize_by):
    if normalize_by == 'median':
        norm_function = np.median
    elif normalize_by == 'max':
        norm_function = np.max
    elif normalize_by == 'mean':
        norm_function = np.mean
    else:
        raise ValueError("The method to normalize by is not recognized. "\
                       + "Please choose either 'mean', 'median', or 'max'.")

    dim_t, num_channels = asig.shape
    norm_asig = asig.as_array()
    for i in range(num_channels):
        norm_value = norm_function(norm_asig[:,i])
        if norm_value:
            norm_asig[:,i] /= norm_value
        else:
            print("Normalization factor is {} for channel {} "\
                  .format(norm_value, i) + "and was skipped.")

    new_asig = asig.duplicate_with_new_data(norm_asig, units='dimensionless')
    new_asig.array_annotations = asig.array_annotations
    return new_asig


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = normalize(block.segments[0].analogsignals[0], args.normalize_by)

    asig.description += "Normalized by {} ({})."\
                        .format(args.normalize_by, os.path.basename(__file__))
    block.segments[0].analogsignals[0] = asig

    write_neo(args.output, block)



import re
import sys
import warnings
from pathlib import Path

import numpy as np

from .io_utils import load_neo


def get_base_type(datatype):
    if hasattr(datatype, "dtype"):
        datatype = datatype.dtype
    elif not type(datatype) == type:
        datatype = type(datatype)

    if datatype == list:
        warnings.warn("List don't have a defined type!")

    if np.issubdtype(datatype, np.integer):
        return "int"
    elif np.issubdtype(datatype, float):
        return "float"
    elif np.issubdtype(datatype, str):
        return "str"
    elif np.issubdtype(datatype, complex):
        return "complex"
    elif np.issubdtype(datatype, bool):
        return "bool"
    else:
        warnings.warn(f"Did not recognize type {datatype}! returning 'object'")
    return "object"


nan_values = {
    "int": -1,
    "float": np.nan,
    "bool": False,
    "a5": "None",
    "str": "None",
    "complex": np.nan + 1j * np.nan,
    "object": None,
}


def get_nan_value(type_string):
    return nan_values[type_string]


def guess_type(string):
    try:
        out = int(string)
    except:
        try:
            out = float(string)
        except:
            out = str(string)
            if out == "None":
                out = None
            elif out == "True":
                out = True
            elif out == "False":
                out = False
    return out


def str2dict(string):
    """
    Transforms a str(dict) back to dict
    """
    if string[0] == "{":
        string = string[1:]
    if string[-1] == "}":
        string = string[:-1]
    my_dict = {}
    # list or tuple values
    brackets = [delimiter for delimiter in ["[", "]", "(", ")"] if delimiter in string]
    if len(brackets):
        for kv in string.split("{},".format(brackets[1])):
            k, v = kv.split(":")
            v = v.replace(brackets[0], "").replace(brackets[1], "")
            values = [guess_type(val) for val in v.split(",")]
            if len(values) == 1:
                values = values[0]
            my_dict[k.strip()] = values
    # scalar values
    else:
        for kv in string.split(","):
            k, v = kv.split(":")
            my_dict[k.strip()] = guess_type(v.strip())
    return my_dict


def parse_string2dict(kwargs_str, **kwargs):
    if type(kwargs_str) == list:
        if len(kwargs_str) == 0:
            return {}
        elif len(kwargs_str) == 1:
            kwargs = kwargs_str[0]
        else:
            kwargs = "".join(kwargs_str)[1:-1]
    else:
        kwargs = str(kwargs_str)
    if guess_type(kwargs) is None:
        return {}

    my_dict = {}
    # match all nested dicts
    pattern = re.compile("[\w\s]+:{[^}]*},*")
    for match in pattern.findall(kwargs):
        nested_dict_name, nested_dict = match.split(":{")
        nested_dict = nested_dict[:-1]
        my_dict[nested_dict_name] = str2dict(nested_dict)
        kwargs = kwargs.replace(match, "")
    # match entries with word value, list value, or tuple value
    pattern = re.compile("[\w\s]+:(?:[\w\.\s\/\-\&\+]+|\[[^\]]+\]|\([^\)]+\))")
    for match in pattern.findall(kwargs):
        my_dict.update(str2dict(match))
    return my_dict


# def ordereddict_to_dict(input_dict):
#     if isinstance(input_dict, dict):
#         for k, v in input_dict.items():
#             if isinstance(v, dict):
#                 input_dict[k] = ordereddict_to_dict(v)
#         return dict(input_dict)
#     else:
#         return input_dict

_true_set = {"yes", "true", "t", "y", "1"}
_false_set = {"no", "false", "f", "n", "0"}


def str_to_bool(value, raise_exc=False):
    if isinstance(value, str):
        value = value.lower()
        if value in _true_set:
            return True
        if value in _false_set:
            return False
    if raise_exc:
        raise ValueError('Expected "%s"' % '", "'.join(_true_set | _false_set))
    return None


def none_or_X(value, dtype):
    if value is None or not bool(value) or value == "None":
        return None
    try:
        return dtype(value)
    except ValueError:
        return None


none_or_int = lambda v: none_or_X(v, int)
none_or_float = lambda v: none_or_X(v, float)
none_or_str = lambda v: none_or_X(v, str)
str_list = lambda v: v.split(",")


def parse_plot_channels(channels, input_file):
    channels = channels if isinstance(channels, list) else [channels]
    channels = [none_or_int(channel) for channel in channels]
    # ToDo:
    #   * check is channel exists, even when there is no None
    #   * use annotation channel ids instead of array indices
    if None in channels:
        dim_t, channel_num = load_neo(
            input_file, object="analogsignal", lazy=True
        ).shape
        for i, channel in enumerate(channels):
            if channel is None or channel >= channel_num:
                channels[i] = np.random.randint(0, channel_num)
    return channels


def determine_spatial_scale(coords):
    coords = np.array(coords)
    dists = np.diff(coords[:, 0])
    dists = dists[np.nonzero(dists)]
    return np.min(dists)



"""
Calculate the spatial derivative on the time-delays of the wave triggers 
in each channel.

The derivative is calculated using a kernel convolution.
"""

import argparse
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.interpolate import RBFInterpolator
from utils.convolve import get_kernel, nan_conv2d
from utils.io_utils import load_neo, save_plot
from utils.parse import none_or_str, str_to_bool

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--output_img", nargs='?', type=none_or_str, default=None,
                 help="path of output image file")
CLI.add_argument("--kernel", "--KERNEL", nargs='?', type=none_or_str, default=None,
                 help="derivative kernel")
CLI.add_argument("--event_name", "--EVENT_NAME", nargs='?', type=str, default='wavefronts',
                 help="name of neo.Event to analyze (must contain waves)")
CLI.add_argument("--interpolate", "--INTERPOLATE", nargs='?', type=str_to_bool, default=False,
                 help="whether to thin-plate-spline interpolate the wave patterns before derivation")
CLI.add_argument("--smoothing", "--SMOOTHING", nargs='?', type=float, default=0,
                 help="smoothing factor for the interpolation")

def interpolate_grid(grid, smoothing):
    y, x = np.where(np.isfinite(grid))
    rbf_func = RBFInterpolator(np.stack((y,x), axis=-1),
                               grid[y,x],
                               neighbors=None,
                               smoothing=smoothing,
                               kernel='thin_plate_spline',
                               epsilon=None,
                               degree=None)
    return rbf_func

def sample_wave_pattern(pattern_func, dim_x, dim_y):
    fy, fx = np.meshgrid(np.arange(0, dim_y),
                         np.arange(0, dim_x),
                         indexing='ij')
    fcoords = np.stack((fy,fx), axis=-1)
    wave_pattern = pattern_func(fcoords.reshape(-1,2))
    return wave_pattern.reshape(dim_y, dim_x)

def calc_spatial_derivative(evts, kernel_name, interpolate=False, smoothing=0):
    labels = evts.labels.astype(int)
    dim_x = int(max(evts.array_annotations['x_coords']))+1
    dim_y = int(max(evts.array_annotations['y_coords']))+1

    spatial_derivative_df = pd.DataFrame()

    for wave_id in np.unique(labels):
        wave_trigger_evts = evts[labels == wave_id]

        x_coords = wave_trigger_evts.array_annotations['x_coords'].astype(int)
        y_coords = wave_trigger_evts.array_annotations['y_coords'].astype(int)
        channels = wave_trigger_evts.array_annotations['channels'].astype(int)

        trigger_collection = np.empty([dim_y, dim_x]) * np.nan
        trigger_collection[y_coords, x_coords] = wave_trigger_evts.times

        if interpolate:
            try:
                pattern_func = interpolate_grid(trigger_collection, smoothing)
                trigger_collection = sample_wave_pattern(pattern_func,
                                                         dim_x=dim_x, dim_y=dim_y)
            except ValueError as ve:
                warn(repr(ve))
                warn('Continuing without interpolation.')

        kernel = get_kernel(kernel_name)
        d_horizont = -1 * nan_conv2d(trigger_collection, kernel.x) # is -1 correct?
        d_vertical = -1 * nan_conv2d(trigger_collection, kernel.y)

        dt_y = d_vertical[y_coords, x_coords]
        dt_x = d_horizont[y_coords, x_coords]

        df = pd.DataFrame(list(zip(dt_x, dt_y, x_coords, y_coords, channels)),
                          columns=['dt_x', 'dt_y', 'x_coords', 'y_coords', 'channel_id'])
        df[f'{evts.name}_id'] = wave_id

        spatial_derivative_df = pd.concat([spatial_derivative_df, df],
                                          ignore_index=True)

    fig, ax = plt.subplots(ncols=3, figsize=(15,5))
    img = ax[0].imshow(trigger_collection, cmap='viridis', origin='lower')
    plt.colorbar(img, ax=ax[0])
    vminmax = np.nanmax(abs(d_vertical))
    img = ax[1].imshow(d_vertical, origin='lower', cmap='coolwarm',
                       vmin=-vminmax, vmax=vminmax)
    plt.colorbar(img, ax=ax[1])
    vminmax = np.nanmax(abs(d_horizont))
    img = ax[2].imshow(d_horizont, origin='lower', cmap='coolwarm',
                       vmin=-vminmax, vmax=vminmax)
    plt.colorbar(img, ax=ax[2])
    ax[0].set_title(f'wave {wave_id}')
    ax[1].set_title('dt Y (vertical)')
    ax[2].set_title('dt X (horizontal)')
    ax[0].set_axis_off()
    ax[1].set_axis_off()
    ax[2].set_axis_off()

    return spatial_derivative_df


if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    asig = block.segments[0].analogsignals[0]
    evts = block.filter(name=args.event_name, objects="Event")[0]
    evts = evts[evts.labels != '-1']

    df = calc_spatial_derivative(evts,
                                 kernel_name=args.kernel,
                                 interpolate=args.interpolate,
                                 smoothing=args.smoothing)

    df['kernel'] = args.kernel
    df['spatial_scale'] = evts.annotations['spatial_scale'].magnitude
    df['spatial_scale_unit'] = evts.annotations['spatial_scale'].dimensionality.string
    df['dt_unit'] = evts.times.dimensionality.string

    df.to_csv(args.output)

    if args.output_img is not None:
        save_plot(args.output_img)



"""
Check whether the input data representation adheres to the stage's requirements.

Additionally prints a short summary of the data attributes.
"""

import numpy as np
import argparse
import quantities as pq
from utils.io_utils import load_neo
from snakemake.logging import logger

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    if len(block.segments) > 1:
        print("More than one Segment found; all except the first one " \
            + "will be ignored.")
    if len(block.segments[0].analogsignals) > 1:
        print("More than one AnalogSignal found; all except the first one " \
            + "will be ignored.")

    asig = block.segments[0].analogsignals[0]

    # print('Recording Time:\t\t', asig.t_stop - asig.t_start)
    # print('Sampling Rate:\t\t', asig.sampling_rate)
    # print('Spatial Scale:\t\t', asig.annotations['spatial_scale'])

    evts = block.filter(name='transitions', objects="Event")

    if not len(evts):
        raise ValueError("No 'transitions' events found!")
    evt = evts[0]

    if not 'UP' in evt.labels:
        logger.warning("No transitions labeled 'UP' found!")
        # raise KeyError("No transitions labeled 'UP' found!")

    up_channels = np.unique(evt.array_annotations['channels'])
    num_channels = np.count_nonzero(~np.isnan(np.sum(asig, axis=0)))
    print(f'{len(up_channels)} of {num_channels} channels show UP transitions.')

    evt.array_annotations['x_coords']
    evt.array_annotations['y_coords']



"""
Cut data according to a start and stop time.
"""

import argparse
import quantities as pq
from utils.io_utils import load_neo, write_neo
from utils.neo_utils import time_slice
from utils.parse import none_or_float

CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=str, required=True,
                 help="path to input data in neo format")
CLI.add_argument("--output", nargs='?', type=str, required=True,
                 help="path of output file")
CLI.add_argument("--t_start", nargs='?', type=none_or_float, default=0,
                 help="new starting time in s")
CLI.add_argument("--t_stop", nargs='?', type=none_or_float, default=10,
                 help="new stopping time in s")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    block = load_neo(args.data)

    for i, asig in enumerate(block.segments[0].analogsignals):
        block.segments[0].analogsignals[i] = time_slice(asig,
                                                        t_start=args.t_start,
                                                        t_stop=args.t_stop)

    for i, evt in enumerate(block.segments[0].events):
        block.segments[0].events[i] = time_slice(evt,
                                                 t_start=args.t_start,
                                                 t_stop=args.t_stop)


    write_neo(args.output, block)



"""
Loads a dataset and brings it into the required data representation (using Neo).
"""

import argparse
import quantities as pq
from pathlib import Path
import neo
from utils.parse import parse_string2dict, none_or_float, none_or_int, none_or_str
from utils.neo_utils import imagesequence_to_analogsignal, merge_analogsignals
from utils.neo_utils import flip_image, rotate_image, time_slice
from utils.io_utils import load_neo, write_neo


CLI = argparse.ArgumentParser()
CLI.add_argument("--data", nargs='?', type=Path, required=True,
                 help="path to input data")
CLI.add_argument("--output", nargs='?', type=Path, required=True,
                 help="path of output file")
CLI.add_argument("--data_name", nargs='?', type=str, default='None',
                 help="chosen name of the dataset")
CLI.add_argument("--sampling_rate", nargs='?', type=none_or_float,
                 default=None, help="sampling rate in Hz")
CLI.add_argument("--spatial_scale", nargs='?', type=float, required=True,
                 help="distance between electrodes or pixels in mm")
CLI.add_argument("--t_start", nargs='?', type=none_or_float, default=None,
                 help="start time, in s, delimits the interval of recordings to be analyzed")
CLI.add_argument("--t_stop", nargs='?', type=none_or_float, default=None,
                 help="stop time, in s, delimits the interval of recordings to be analyzed")
CLI.add_argument("--orientation_top", nargs='?', type=str, required=True,
                 help="upward orientation of the recorded cortical region")
CLI.add_argument("--orientation_right", nargs='?', type=str, required=True,
                 help="right-facing orientation of the recorded cortical region")
CLI.add_argument("--annotations", nargs='+', type=none_or_str, default=None,
                 help="metadata of the dataset")
CLI.add_argument("--array_annotations", nargs='+', type=none_or_str,
                 default=None, help="channel-wise metadata")
CLI.add_argument("--kwargs", nargs='+', type=none_or_str, default=None,
                 help="additional optional arguments")

if __name__ == '__main__':
    args, unknown = CLI.parse_known_args()

    # Load data with Neo IO or custom loading routine
    block = load_neo(args.data)
    # If there is no Neo IO for the data type available,
    # the data must be loaded conventionally and added to a newly constructed
    # Neo block. For building a Neo objects, have a look into the documentation
    # https://neo.readthedocs.io/

    # In case the dataset is imaging data and therefore stored as an
    # ImageSequence object, it needs to be transformed into an AnalogSignal
    # object. To do this use the function imagesequence_to_analogsignal in utils/neo_utils.py

    asig = block.segments[0].analogsignals[0]

    asig = time_slice(asig, args.t_start, args.t_stop)

    # Add metadata from ANNOTATION dict
    asig.annotations.update(parse_string2dict(args.annotations))
    asig.annotations.update(spatial_scale=args.spatial_scale*pq.mm)
    asig.annotations.update(orientation_top=args.orientation_top)
    asig.annotations.update(orientation_right=args.orientation_right)

    # Add metadata from ARRAY_ANNOTATION dict
    asig.array_annotations.update(parse_string2dict(args.array_annotations))

    # Do custom metadata processing from KWARGS dict (optional)
    # kwargs = parse_string2dict(args.kwargs)
    # ... do something

    # Add description to the Neo object
    block.name = args.data_name
    block.segments[0].name = 'Segment 1'
    block.segments[0].description = 'Loaded with neo.io version {}'\
                                    .format(neo.__version__)
    if asig.description is None:
        asig.description = ''
    asig.description += 'some signal. '

    # Update the annotated AnalogSignal object in the Neo Block
    block.segments[0].analogsignals[0] = asig

    # Save data to file
    write_neo(args.output, block)
